from typing import Iterable, Optional

import numpy as np
import pandas as pd

from ..core.core import ModelType
from ..datasets.base import Dataset
from .base import BaseModel, ModelPredictionResults


class PrecookedModel(BaseModel):
    """A dummy model for internal usage."""

    def __init__(
        self,
        data: Dataset,
        predictions: ModelPredictionResults,
        model_type: ModelType,
        feature_names: Optional[Iterable] = None,
        classification_labels: Optional[Iterable] = None,
        **kwargs,
    ):
        self._data = data
        self._predictions = predictions
        super().__init__(
            model_type=model_type, feature_names=feature_names, classification_labels=classification_labels, **kwargs
        )

    @classmethod
    def from_model(cls, model: BaseModel, dataset: Dataset):
        """Creates a PrecookedModel from an existing model and dataset.

        Parameters
        ----------
        model : BaseModel
            A instance of a Giskard model.
        dataset : Dataset
            Dataset for which predictions will be cached.

        Returns
        -------
        PrecookedModel

        """
        predictions = model.predict(dataset)

        return cls(
            dataset,
            predictions,
            model.meta.model_type,
            model.meta.feature_names,
            model.meta.classification_labels,
        )

    def predict(self, dataset: Dataset, *args, **kwargs) -> ModelPredictionResults:
        refs = pd.Series(np.arange(len(self._data)), index=self._data.df.index)
        idx = refs.loc[dataset.df.index]

        raw = np.asarray(self._predictions.raw)[idx]
        prediction = np.asarray(self._predictions.prediction)[idx]
        raw_prediction = np.asarray(self._predictions.raw_prediction)[idx]

        probabilities = getattr(self._predictions, "probabilities", None)
        if probabilities is not None:
            probabilities = np.asarray(probabilities)[idx]

        all_predictions = getattr(self._predictions, "all_predictions", None)
        if all_predictions is not None:
            all_predictions = np.asarray(all_predictions)[idx]

        return ModelPredictionResults(
            raw=raw,
            prediction=prediction,
            raw_prediction=raw_prediction,
            probabilities=probabilities,
            all_predictions=all_predictions,
        )

    def predict_df(self, df: pd.DataFrame, *args, **kwargs):
        raise NotImplementedError()



# type: ignore[attr-defined]
"""Inspect your AI models visually, find bugs, give feedback 🕵️‍♀️ 💬"""

# The following code block is made to remove pydantic 2.0 warnings, especially for MLFlow
# https://docs.python.org/3/library/warnings.html#temporarily-suppressing-warnings
from importlib import metadata as importlib_metadata

from . import demo
from .core.suite import Suite, SuiteInput
from .core.test_result import TestResult
from .datasets.base import Dataset
from .models.automodel import Model
from .models.model_explanation import explain_with_shap
from .registry.decorators import test
from .registry.giskard_test import GiskardTest
from .registry.slicing_function import SlicingFunction, slicing_function
from .registry.transformation_function import TransformationFunction, transformation_function
from .scanner import scan
from .utils.analytics_collector import analytics
from .utils.logging_utils import configure_logging

configure_logging()


def get_version() -> str:
    try:
        res = importlib_metadata.version(__name__)
        if res is None:
            # importlib_metadata can return None https://github.com/python/importlib_metadata/issues/371
            # fallback to pkg_resources even if it's deprecated
            import pkg_resources

            return pkg_resources.get_distribution(__name__).version
        return res
    except importlib_metadata.PackageNotFoundError:  # pragma: no cover
        return "unknown"


__version__: str = get_version()

analytics.track("Initialized giskard library")

__all__ = [
    "Dataset",
    "test",
    "Model",
    "Suite",
    "slicing_function",
    "transformation_function",
    "SuiteInput",
    "SlicingFunction",
    "TransformationFunction",
    "scan",
    "explain_with_shap",
    "TestResult",
    "GiskardTest",
    "demo",
]



from typing import TYPE_CHECKING, Any, Dict, Optional, Sequence

import json
import logging
from dataclasses import dataclass
from pathlib import Path

import pandas as pd
from datasets import Dataset as HFDataset
from datasets import load_dataset
from huggingface_hub import DatasetCard

if TYPE_CHECKING:
    from huggingface_hub import CommitInfo

from ..core.suite import Suite
from ..datasets.base import Dataset
from ..testing.tests.llm import test_llm_correctness

logger = logging.getLogger(__name__)


@dataclass
class QuestionSample:
    id: str
    question: str
    reference_answer: str
    reference_context: str
    conversation_history: Sequence[Dict[str, str]]
    metadata: Dict[str, Any]
    agent_answer: Optional[str] = None
    correctness: Optional[bool] = None

    def to_dict(self):
        sample_dict = {
            "id": self.id,
            "question": self.question,
            "reference_answer": self.reference_answer,
            "reference_context": self.reference_context,
            "conversation_history": self.conversation_history,
            "metadata": self.metadata,
        }
        if self.agent_answer is not None:
            sample_dict["agent_answer"] = self.agent_answer
        if self.correctness is not None:
            sample_dict["correctness"] = self.correctness
        return sample_dict


class QATestset:
    """A class to represent a testset for QA models."""

    def __init__(self, question: Sequence[QuestionSample]):
        self._questions = question
        self._dataframe = pd.DataFrame.from_records([question.to_dict() for question in self._questions]).set_index(
            "id"
        )

    def __len__(self):
        return len(self._dataframe)

    @classmethod
    def from_pandas(cls, dataframe: pd.DataFrame):
        return cls([QuestionSample(**record) for record in dataframe.to_dict(orient="records")])

    def to_pandas(self, filters: Optional[dict] = None):
        """Return the testset as a pandas DataFrame.

        Parameters
        ----------
        filters : dict, optional
            A dictionary of filters to apply to the testset. The keys are the metadata names and the values are list of kept values.
            If multiple metadata are provided, the filters are combined with an AND operator.

        """

        if filters is not None:
            return self._dataframe[
                self._dataframe["metadata"].apply(lambda x: all(x.get(k) in v for k, v in filters.items()))
            ]
        return self._dataframe

    def to_dataset(self, filters: Optional[dict] = None):
        dataframe = self.to_pandas(filters=filters).copy()
        dataframe["conversation_history"] = dataframe["conversation_history"].apply(lambda x: json.dumps(x))
        dataframe["metadata"] = dataframe["metadata"].apply(lambda x: json.dumps(x))
        return Dataset(dataframe, name="QA Testset", target=False, validation=False)

    def get_metadata_values(self, metadata_name: str):
        """Return the unique values of a metadata field."""
        metadata_values = list(
            set([metadata[metadata_name] for metadata in self._dataframe["metadata"] if metadata_name in metadata])
        )
        metadata_values.sort()
        return metadata_values

    def count_sample_by_metadata(self, metadata_name: str):
        """Count the number of samples for each value of a metadata field."""
        return self._dataframe["metadata"].apply(lambda x: x.get(metadata_name)).value_counts()

    def save(self, path):
        """Save the testset as a JSONL file.

        Parameters
        ----------
        path : str
            The path to the output JSONL file.
        """
        self._dataframe.reset_index().to_json(path, orient="records", lines=True, force_ascii=False)

    @classmethod
    def load(cls, path):
        """Load a testset from a JSONL file.

        Parameters
        ----------
        path : str
            The path to the input JSONL file.
        """
        dataframe = pd.read_json(path, orient="records", lines=True)
        return cls.from_pandas(dataframe)

    def push_to_hf_hub(
        self,
        repo_id: str,
        token: str = None,
        private: bool = False,
        **kwargs: Any,
    ) -> "CommitInfo":
        """Push the QATestset to the Hugging Face Hub.

        Parameters
        ----------
        repo_id : str
            The repository ID on the Hugging Face Hub.
        token : str, optional
            Authentication token for private repositories. Defaults to None.
        private : bool
            Whether to create a private repository. Defaults to False.
        **kwargs : Any
            Additional arguments passed to Dataset.push_to_hub().

        Returns
        -------
        CommitInfo
            The commit information.
        """

        # Conversion to Dataset from the datasets library
        dataset = HFDataset.from_pandas(self._dataframe)
        dataset.push_to_hub(repo_id, token=token, private=private, **kwargs)

        # Load the dataset card template
        template_path = Path(__file__).parent / "dataset_card_template.md"
        template = template_path.read_text()

        # Make and push the dataset card
        try:
            from ..llm.client import get_default_client

            config = {"metadata": get_default_client().get_config()}
        except Exception:
            config = {}
        content = template.format(repo_id=repo_id, num_items=len(self._dataframe), config=json.dumps(config, indent=4))
        return DatasetCard(content=content).push_to_hub(repo_id=repo_id, token=token, repo_type="dataset")

    @classmethod
    def load_from_hf_hub(cls, repo_id: str, token: str = None, **kwargs: Any) -> "QATestset":
        """
        Load an instance of the class from the Hugging Face Hub.

        Parameters
        ----------
        repo_id : str
            The repository ID on the Hugging Face Hub.
        token : str, optional
            Authentication token for private repositories. Defaults to None.
        **kwargs : Any
            Additional arguments passed to `load_dataset`.

        Returns
        -------
        QATestset
            An instance of the class itself loaded from the Hub.

        Raises
        ------
        ImportError
            If required dependencies are not installed.
        """
        dataset = load_dataset(repo_id, token=token, split="train", **kwargs)
        dataframe = pd.DataFrame(dataset)
        return cls.from_pandas(dataframe)

    def to_test_suite(self, name=None, slicing_metadata: Optional[Sequence[str]] = None):
        """
        Convert the testset to a Giskard test suite.

        Parameters
        ----------
        name : str, optional
            The name of the test suite. If not provided, the name will be "Test suite generated from testset".

        Returns
        -------
        giskard.Suite
            The test suite.
        """

        if slicing_metadata is None:
            suite_default_params = {"dataset": self.to_dataset()}
            name = name or "Test suite generated from testset"
            suite = Suite(name=name, default_params=suite_default_params)
            suite.add_test(test_llm_correctness, "TestsetCorrectnessTest", "TestsetCorrectnessTest")
        else:
            name = name or "Test suite generated from testset"
            suite = Suite(name=name)
            for metadata_name in slicing_metadata:
                metadata_values = self.get_metadata_values(metadata_name)
                for value in metadata_values:
                    suite.add_test(
                        test_llm_correctness(dataset=self.to_dataset(filters={metadata_name: [value]})),
                        f"TestsetCorrectnessTest_{metadata_name}_{value}",
                        f"TestsetCorrectnessTest_{metadata_name}_{value}",
                    )
        return suite

    def copy(self):
        """Return a copy of the testset."""
        return QATestset(self._dataframe.copy())

    @property
    def samples(self):
        return self._questions



from ..knowledge_base import KnowledgeBase
from ..testset import QuestionSample
from .base import _BaseModifierGenerator
from .prompt import QAGenerationPrompt
from .simple_questions import SimpleQuestionsGenerator

DISTRACTING_SYSTEM_PROMPT = """You are an expert at rewriting questions.
Your task is to re-write questions that will be used to evaluate the following agent:
- Agent description: {agent_description}  

Your task is to complexify questions given a provided context. 
Please respect the following rules to generate the question:
- The new question must include a condition or constraint based on the provided context. 
- The original question direction should be preserved.
- The question must be plausible according to the context and the agent description.
- The question must be self-contained and understandable by humans. 
- The question must be in this language: {language}

You will be provided the question delimited with <question></question> tags.
You will also be provided a context paragraph delimited with <context></context> tags.
You will return the reformulated question as a single JSON object, with the key 'question'. Make sure you return a valid JSON object.
"""

DISTRACTING_INPUT_TEMPLATE = """<question>
{question}
</question>
<answer>
{answer}
</answer>
<context>
{context}
</context>"""

DISTRACTING_EXAMPLE_INPUT = DISTRACTING_INPUT_TEMPLATE.format(
    question="What job offer do you have for engineering student?",
    answer="We have plenty of different jobs for engineering student depending on your speciality: mechanical engineer, data scientist, electronic designer and many more.",
    context="Sometimes employers assume being accessible and inclusive only means providing physical access like ramps, accessible bathrooms and automatic opening doors. However, there are many other important ways to demonstrate that you welcome and want to attract a diverse workforce including people with disability.",
)

DISTRACTING_EXAMPLE_OUTPUT = """{
    "question": "Do you have any job opening suitable for engineering students with a disability? "
}"""


class DistractingQuestionsGenerator(_BaseModifierGenerator):
    _base_generator = SimpleQuestionsGenerator()

    _prompt = QAGenerationPrompt(
        system_prompt=DISTRACTING_SYSTEM_PROMPT,
        example_input=DISTRACTING_EXAMPLE_INPUT,
        example_output=DISTRACTING_EXAMPLE_OUTPUT,
        user_input_template=DISTRACTING_INPUT_TEMPLATE,
    )

    _question_type = "distracting element"

    def _modify_question(
        self, question: QuestionSample, knowledge_base: KnowledgeBase, agent_description: str, language: str
    ) -> QuestionSample:
        distracting_context = knowledge_base.get_random_document().content
        messages = self._prompt.to_messages(
            system_prompt_input={
                "agent_description": agent_description,
                "language": language,
            },
            user_input={
                "question": question.question,
                "answer": question.reference_answer,
                "context": distracting_context,
            },
        )
        question.metadata["question_type"] = self._question_type
        question.metadata["distracting_context"] = distracting_context
        out = self._llm_complete(messages=messages)
        question.question = out["question"]
        return question


distracting_questions = DistractingQuestionsGenerator()



from .adversarial import AdversarialDataGenerator
from .base import LLMBasedDataGenerator
from .implausible import ImplausibleDataGenerator
from .sycophancy import SycophancyDataGenerator

__all__ = [
    "LLMBasedDataGenerator",
    "SycophancyDataGenerator",
    "ImplausibleDataGenerator",
    "AdversarialDataGenerator",
]



from typing import Sequence

from ..llm.client.base import ChatMessage, LLMClient

RECOMMENDATION_SYSTEM_PROMPT = """You are an expert at designing Retrieval Augmented Generation (RAG) systems. You new task is to help debug a RAG based on several scores.

The RAG is composed of the following components:
- Knowledge base: contains all the documents and information the RAG has access to.
- Retriever: an algorithm that find relevant documents in the knowledge base, given a user's question.
- Generator: a LLM model that generates the answers to the user's questions.
- Rewriter: rewrites the user questions to match better with the use case or the knowledge base.
- Router: a filter that decides which component to use to answer the user's question. It can also be a moderation layer.

The RAG has been evaluated on a set of questions. The questions are of various types:
- Simple questions: questions that can be answered with a single document. (Evaluated components: Generator, Retriever, Router)
- Complex questions: questions more difficult to answer due to paraphrasing. (Evaluated components: Generator)
- Distracting questions: questions which include random information from the knowledge base. (Evaluated components: Generator, Retriever, Rewriter)
- Situational questions: questions that includes situational context given by a user. (Evaluated components: Generator)
- Conversational questions: questions that are part of a conversation. (Evaluated components: Rewriter)
- Double questions: questions that include two questions. (Evaluated components: Rewriter)

While the knowledge base quality is not assessed by any question type, each question is related to exactly one of these topics. 
A proxy for the quality of the knowledge base is the RAG's ability to answer equally well across all topics. Here are the topics:
{topics}

Your are now provided with the scores of the agent on each question type and each topic.
"""

RECOMMENDATION_USER_PROMPT = """
Scores per question type:
{scores_per_question_type}

Scores per topic:
{scores_per_topic}

Give me a short recommendation including the most important action I should take to improve my RAG system.
Your answer should be at most 2 sentences long.
"""


def get_rag_recommendation(
    topics: Sequence[str], scores_per_question_type: dict, scores_per_topic: dict, llm_client: LLMClient
):
    formatted_topics = "\n".join([f"- {topic}" for idx, topic in enumerate(topics)])
    formatted_scores_per_question_type = "\n".join(
        [f"- {question_type}: {score}" for question_type, score in scores_per_question_type.items()]
    )
    formatted_scores_per_topic = "\n".join([f"- {topic}: {score}" for topic, score in scores_per_topic.items()])

    recommendation = llm_client.complete(
        [
            ChatMessage(role="system", content=RECOMMENDATION_SYSTEM_PROMPT.format(topics=formatted_topics)),
            ChatMessage(
                role="user",
                content=RECOMMENDATION_USER_PROMPT.format(
                    scores_per_question_type=formatted_scores_per_question_type,
                    scores_per_topic=formatted_scores_per_topic,
                ),
            ),
        ]
    )
    return recommendation.content



import random
import re

import pandas as pd
from scipy.stats import median_abs_deviation

from ..datasets import Dataset
from ..llm import get_default_client
from ..registry.transformation_function import transformation_function

nearbykeys = {
    "a": ["q", "w", "s", "x", "z"],
    "b": ["v", "g", "h", "n"],
    "c": ["x", "d", "f", "v"],
    "d": ["s", "e", "r", "f", "c", "x"],
    "e": ["w", "s", "d", "r"],
    "f": ["d", "r", "t", "g", "v", "c"],
    "g": ["f", "t", "y", "h", "b", "v"],
    "h": ["g", "y", "u", "j", "n", "b"],
    "i": ["u", "j", "k", "o"],
    "j": ["h", "u", "i", "k", "n", "m"],
    "k": ["j", "i", "o", "l", "m"],
    "l": ["k", "o", "p"],
    "m": ["n", "j", "k", "l"],
    "n": ["b", "h", "j", "m"],
    "o": ["i", "k", "l", "p"],
    "p": ["o", "l"],
    "q": ["w", "a", "s"],
    "r": ["e", "d", "f", "t"],
    "s": ["w", "e", "d", "x", "z", "a"],
    "t": ["r", "f", "g", "y"],
    "u": ["y", "h", "j", "i"],
    "v": ["c", "f", "g", "v", "b"],
    "w": ["q", "a", "s", "e"],
    "x": ["z", "s", "d", "c"],
    "y": ["t", "g", "h", "u"],
    "z": ["a", "s", "x"],
}

gruber = re.compile(
    r"""(?i)\b((?:https?:(?:/{1,3}|[a-z0-9%])|[a-z0-9.\-]+[.](?:com|net|org|edu|gov|mil|aero|asia|biz|cat|coop|info|int|jobs|mobi|museum|name|post|pro|tel|travel|xxx|ac|ad|ae|af|ag|ai|al|am|an|ao|aq|ar|as|at|au|aw|ax|az|ba|bb|bd|be|bf|bg|bh|bi|bj|bm|bn|bo|br|bs|bt|bv|bw|by|bz|ca|cc|cd|cf|cg|ch|ci|ck|cl|cm|cn|co|cr|cs|cu|cv|cx|cy|cz|dd|de|dj|dk|dm|do|dz|ec|ee|eg|eh|er|es|et|eu|fi|fj|fk|fm|fo|fr|ga|gb|gd|ge|gf|gg|gh|gi|gl|gm|gn|gp|gq|gr|gs|gt|gu|gw|gy|hk|hm|hn|hr|ht|hu|id|ie|il|im|in|io|iq|ir|is|it|je|jm|jo|jp|ke|kg|kh|ki|km|kn|kp|kr|kw|ky|kz|la|lb|lc|li|lk|lr|ls|lt|lu|lv|ly|ma|mc|md|me|mg|mh|mk|ml|mm|mn|mo|mp|mq|mr|ms|mt|mu|mv|mw|mx|my|mz|na|nc|ne|nf|ng|ni|nl|no|np|nr|nu|nz|om|pa|pe|pf|pg|ph|pk|pl|pm|pn|pr|ps|pt|pw|py|qa|re|ro|rs|ru|rw|sa|sb|sc|sd|se|sg|sh|si|sj|Ja|sk|sl|sm|sn|so|sr|ss|st|su|sv|sx|sy|sz|tc|td|tf|tg|th|tj|tk|tl|tm|tn|to|tp|tr|tt|tv|tw|tz|ua|ug|uk|us|uy|uz|va|vc|ve|vg|vi|vn|vu|wf|ws|ye|yt|yu|za|zm|zw)/)(?:[^\s()<>{}\[\]]|\([^\s()]*?\([^\s()]+\)[^\s()]*?\)|\([^\s]+?\))+(?:\([^\s()]*?\([^\s()]+\)[^\s()]*?\)|\([^\s]+?\)|[^\s`!()\[\]{};:'".,<>?«»“”‘’])|(?:(?<!@)[a-z0-9]+(?:[.\-][a-z0-9]+)*[.](?:com|net|org|edu|gov|mil|aero|asia|biz|cat|coop|info|int|jobs|mobi|museum|name|post|pro|tel|travel|xxx|ac|ad|ae|af|ag|ai|al|am|an|ao|aq|ar|as|at|au|aw|ax|az|ba|bb|bd|be|bf|bg|bh|bi|bj|bm|bn|bo|br|bs|bt|bv|bw|by|bz|ca|cc|cd|cf|cg|ch|ci|ck|cl|cm|cn|co|cr|cs|cu|cv|cx|cy|cz|dd|de|dj|dk|dm|do|dz|ec|ee|eg|eh|er|es|et|eu|fi|fj|fk|fm|fo|fr|ga|gb|gd|ge|gf|gg|gh|gi|gl|gm|gn|gp|gq|gr|gs|gt|gu|gw|gy|hk|hm|hn|hr|ht|hu|id|ie|il|im|in|io|iq|ir|is|it|je|jm|jo|jp|ke|kg|kh|ki|km|kn|kp|kr|kw|ky|kz|la|lb|lc|li|lk|lr|ls|lt|lu|lv|ly|ma|mc|md|me|mg|mh|mk|ml|mm|mn|mo|mp|mq|mr|ms|mt|mu|mv|mw|mx|my|mz|na|nc|ne|nf|ng|ni|nl|no|np|nr|nu|nz|om|pa|pe|pf|pg|ph|pk|pl|pm|pn|pr|ps|pt|pw|py|qa|re|ro|rs|ru|rw|sa|sb|sc|sd|se|sg|sh|si|sj|Ja|sk|sl|sm|sn|so|sr|ss|st|su|sv|sx|sy|sz|tc|td|tf|tg|th|tj|tk|tl|tm|tn|to|tp|tr|tt|tv|tw|tz|ua|ug|uk|us|uy|uz|va|vc|ve|vg|vi|vn|vu|wf|ws|ye|yt|yu|za|zm|zw)\b/?(?!@)))"""
)  # noqa


@transformation_function(name="Keyboard typo", tags=["text"], cell_level=True)
def keyboard_typo_transformation(text: str, rate: float = 0.1) -> str:
    """
    Generate a random typo from words of the text of 'column_name'
    Typos are generated through character substitution based on keyboard proximity
    """
    # Split the text into words
    if pd.isnull(text):
        return text

    words = text.split(" ")

    # Introduce typos into some of the words
    for i in range(len(words)):
        if random.random() < rate:
            word = words[i]
            if len(word) > 1:
                j = random.randint(0, len(word) - 1)
                c = word[j]
                if c in nearbykeys:
                    replacement = random.choice(nearbykeys[c])
                    words[i] = word[:j] + replacement + word[j + 1 :]

    # Join the words back into a string
    return " ".join(words)


@transformation_function(name="Transform to uppercase", row_level=False)
def text_uppercase(data: pd.DataFrame, column: str):
    from ..scanner.robustness.text_transformations import TextUppercase

    return TextUppercase(column).execute(data)


@transformation_function(name="Transform to lowercase", row_level=False)
def text_lowercase(data: pd.DataFrame, column: str):
    from ..scanner.robustness.text_transformations import TextLowercase

    return TextLowercase(column).execute(data)


@transformation_function(name="Transform to title case", row_level=False)
def text_title_case(data: pd.DataFrame, column: str):
    from ..scanner.robustness.text_transformations import TextTitleCase

    return TextTitleCase(column).execute(data)


@transformation_function(name="Add typos", row_level=False)
def text_typo(data: pd.DataFrame, column: str, rate: float = 0.05, min_length: int = 10, rng_seed: int = 1729):
    from ..scanner.robustness.text_transformations import TextTypoTransformation

    return TextTypoTransformation(column, rate, min_length, rng_seed).execute(data)


@transformation_function(name="Add typos from OCR", row_level=False)
def text_typo_from_ocr(data: pd.DataFrame, column: str, rate: float = 0.05, min_length: int = 10, rng_seed: int = 1729):
    from ..scanner.robustness.text_transformations import TextFromOCRTypoTransformation

    return TextFromOCRTypoTransformation(column, rate, min_length, rng_seed).execute(data)


@transformation_function(name="Punctuation Removal", row_level=False)
def text_punctuation_removal(data: pd.DataFrame, column: str):
    from ..scanner.robustness.text_transformations import TextPunctuationRemovalTransformation

    return TextPunctuationRemovalTransformation(column).execute(data)


@transformation_function(name="Accent Removal", row_level=False)
def text_accent_removal(data: pd.DataFrame, column: str, rate: float = 1.0, rng_seed: int = 1729):
    from ..scanner.robustness.text_transformations import TextAccentRemovalTransformation

    return TextAccentRemovalTransformation(column, rate, rng_seed).execute(data)


@transformation_function(name="Switch Gender", row_level=False, needs_dataset=True)
def text_gender_switch(dataset: Dataset, column: str):
    from ..scanner.robustness.text_transformations import TextGenderTransformation

    return TextGenderTransformation(column).execute(dataset)


@transformation_function(name="Transform numbers to words", row_level=False, needs_dataset=True)
def text_number_to_word(dataset: Dataset, column: str):
    from ..scanner.robustness.text_transformations import TextNumberToWordTransformation

    return TextNumberToWordTransformation(column).execute(dataset)


@transformation_function(name="Switch Religion", row_level=False, needs_dataset=True)
def text_religion_switch(dataset: Dataset, column: str):
    from ..scanner.robustness.text_transformations import TextReligionTransformation

    return TextReligionTransformation(column).execute(dataset)


@transformation_function(
    name="Switch countries from high- to low-income and vice versa", row_level=False, needs_dataset=True
)
def text_nationality_switch(dataset: Dataset, column: str):
    from ..scanner.robustness.text_transformations import TextNationalityTransformation

    return TextNationalityTransformation(column).execute(dataset)


@transformation_function(name="Add text from speech typos", row_level=False, needs_dataset=True)
def text_typo_from_speech(dataset: Dataset, column: str, rng_seed: int = 1729, min_length: int = 10):
    from ..scanner.robustness.text_transformations import TextFromSpeechTypoTransformation

    return TextFromSpeechTypoTransformation(column, rng_seed, min_length).execute(dataset)


@transformation_function(name="Change writing style", row_level=False, tags=["text"])
def change_writing_style(df: pd.DataFrame, column_name: str, style: str) -> pd.DataFrame:
    sys_prompt = f"""Your task is to rewrite user-provided text using a given style.

Your goal is to rewrite the provided text according to the specified style. Do not add extra sentences in your answer, only include the rewritten text:
Please ensure that your rewritten text retains the meaning of the original text as much as possible, keep the same
language.

Here is an example:
SYSTEM: Your writing style is: informal
USER: The Third Council of Constantinople, counted as the Sixth Ecumenical Council by the Eastern Orthodox and Catholic Churches, as well as by certain other Western Churches, met in 680–681 and condemned monoenergism and monothelitism as heretical and defined Jesus Christ as having two energies and two wills (divine and human).
ASSISTANT: So, there was this big meeting called the Third Council of Constantinople, which some churches refer to as the Sixth Ecumenical Council. It happened in 680–681, and they basically said that monoenergism and monothelitism were heretical. They also made it clear that Jesus Christ has two energies and two wills, one divine and one human.

Now it’s your turn. Your writing style is: {style}
"""

    client = get_default_client()

    def _rewrite(text):
        llm_output = client.complete(
            messages=[
                {"role": "system", "content": sys_prompt},
                {"role": "user", "content": text},
            ],
            model="gpt-3.5-turbo",
            caller_id="change_writing_style_transormation",
        )
        return llm_output.message

    new_df = df.copy()
    new_df[column_name] = df[column_name].apply(_rewrite)

    return new_df


def compute_mad(x):
    return median_abs_deviation(x, scale=1)


@transformation_function(name="MAD Increment", tags=["num"], row_level=False)
def mad_transformation(
    data: pd.DataFrame, column_name: str, factor: float = 1, value_added: float = None
) -> pd.DataFrame:
    """
    Add 3 times the value_added to the column, or if unavailable, add 3 times the MAD value.
    """
    data = data.copy()
    if value_added is None:
        value_added = compute_mad(data[column_name])
    data[column_name] = data[column_name].apply(lambda x: x + factor * value_added).astype(data[column_name].dtype)
    return data


@transformation_function(name="Add value", tags=["numerical"], row_level=False)
def add_value(data: pd.DataFrame, column_name: str, value_added: float = 0.0) -> pd.DataFrame:
    """
    Add the value_added to the column.
    """
    data = data.copy()
    data[column_name] = data[column_name].apply(lambda x: x + value_added).astype(data[column_name].dtype)
    return data



from typing import Dict, Iterator, Optional, Sequence

import logging
from abc import ABC, abstractmethod

from ...llm.client import ChatMessage, LLMClient, get_default_client
from ..knowledge_base import KnowledgeBase
from .utils import parse_json_output

logger = logging.getLogger("giskard.rag")


class QuestionGenerator(ABC):
    @abstractmethod
    def generate_questions(self, knowledge_base: KnowledgeBase, num_questions: int, *args, **kwargs) -> Iterator[int]:
        ...


class _LLMBasedQuestionGenerator(QuestionGenerator):
    def __init__(
        self,
        context_neighbors: int = 4,
        context_similarity_threshold: float = 0.2,
        context_window_length: int = 8192,
        llm_client: Optional[LLMClient] = None,
        llm_temperature: float = 0.4,
    ):
        self._context_window_length = context_window_length
        self._context_neighbors = context_neighbors
        self._context_similarity_threshold = context_similarity_threshold
        self._llm_client_instance = llm_client
        self._llm_temperature = llm_temperature

    @property
    def _llm_client(self):
        if self._llm_client_instance is None:
            return get_default_client()

        return self._llm_client_instance

    def _llm_complete(self, messages: Sequence[ChatMessage]) -> dict:
        out = self._llm_client.complete(
            messages=messages,
            temperature=self._llm_temperature,
            caller_id=self.__class__.__name__,
        )

        return parse_json_output(out.content, self._llm_client, caller_id=self.__class__.__name__)


class GenerateFromSingleQuestionMixin:
    _question_type: str

    def generate_questions(self, knowledge_base: KnowledgeBase, num_questions: int, *args, **kwargs) -> Iterator[Dict]:
        docs = knowledge_base.get_random_documents(num_questions)

        for doc in docs:
            try:
                yield self.generate_single_question(knowledge_base, *args, **kwargs, seed_document=doc)
            except Exception as e:  # @TODO: specify exceptions
                logger.error(f"Encountered error in question generation: {e}. Skipping.")
                logger.exception(e)


class _BaseModifierGenerator(_LLMBasedQuestionGenerator):
    _base_generator: QuestionGenerator
    _question_type: str

    @abstractmethod
    def _modify_question(initial_question: dict, knowledge_base: KnowledgeBase, *args, **kwargs) -> Dict:
        ...

    def generate_questions(self, knowledge_base: KnowledgeBase, num_questions: int, *args, **kwargs) -> Iterator[Dict]:
        # Generate the base questions
        base_questions = self._base_generator.generate_questions(knowledge_base, num_questions, *args, **kwargs)

        # Run the modifier for each
        for question in base_questions:
            try:
                yield self._modify_question(question, knowledge_base, *args, **kwargs)
            except Exception as e:  # @TODO: specify exceptions
                logger.error(f"Encountered error in question generation: {e}. Skipping.")
                logger.exception(e)



from typing import Sequence

import pandas as pd

from giskard.scanner import logger

from ...datasets.base import Dataset
from ...llm.evaluators import PlausibilityEvaluator
from ...llm.generators import ImplausibleDataGenerator
from ...llm.utils import format_chat_messages
from ...models.base.model import BaseModel
from ...testing.tests.llm.hallucination import test_llm_output_plausibility
from ..decorators import detector
from ..issues import Hallucination, Issue, IssueLevel
from ..registry import Detector
from .base import _estimate_base_token_counts


@detector(
    "llm_implausible_output",
    tags=["hallucination", "misinformation", "implausible_output", "llm", "generative", "text_generation"],
)
class LLMImplausibleOutputDetector(Detector):
    """Detects implausible output in LLM-based models.

    You can explicitly run this detector by adding the tag "implausible_output" in the `only` parameter of the scan method.

    The detector will stimulate the model in producing outputs that are implausible or controversial by generating a
    set of ad hoc adversarial inputs. This can be seen as a proxy for hallucination and misinformation detection.
    """

    def __init__(self, num_samples=10):
        self.num_samples = num_samples

    def get_cost_estimate(self, model: BaseModel, dataset: Dataset) -> dict:
        counts = _estimate_base_token_counts(model, dataset)
        model_meta_tokens = counts["model_meta_tokens"]
        input_sample_tokens = counts["input_sample_tokens"]

        num_calls = 0
        num_prompt_tokens = 0
        num_sampled_tokens = 0

        # Data generation
        num_calls += 1
        num_prompt_tokens += 340 + model_meta_tokens
        num_sampled_tokens += input_sample_tokens * self.num_samples

        # Evaluation, for each generated sample
        num_calls += self.num_samples
        num_prompt_tokens += self.num_samples * (180 + model_meta_tokens + input_sample_tokens + 50)
        num_sampled_tokens += self.num_samples * 15

        return {
            "model_predict_calls": self.num_samples,
            "llm_calls": num_calls,
            "llm_prompt_tokens": num_prompt_tokens,
            "llm_sampled_tokens": num_sampled_tokens,
        }

    def run(self, model: BaseModel, dataset: Dataset, features=None) -> Sequence[Issue]:
        # Generate inputs
        languages = dataset.extract_languages(columns=model.meta.feature_names)

        generator = ImplausibleDataGenerator(llm_temperature=0.1, languages=languages)
        eval_dataset = generator.generate_dataset(
            model, num_samples=self.num_samples, column_types=dataset.column_types
        )
        logger.debug(f"{self.__class__.__name__}: Generated {len(eval_dataset)} inputs")

        # Evaluate the model outputs
        evaluator = PlausibilityEvaluator()
        eval_result = evaluator.evaluate(model, eval_dataset)

        examples = pd.DataFrame(
            [
                {
                    "Conversation": format_chat_messages(r["sample"].get("conversation", [])),
                    "Reason": r.get("reason", "No reason provided."),
                }
                for r in eval_result.failure_examples
            ]
        )

        if eval_result.failed:
            return [
                Issue(
                    model,
                    eval_dataset,
                    group=Hallucination,
                    level=IssueLevel.MEDIUM,
                    description="The model produces implausible output.",
                    meta={
                        "metric": "Failing samples",
                        "metric_value": len(eval_result.failure_examples),
                        "domain": "Implausible or controversial output",
                        "deviation": "The model produces implausible output.",
                        "hide_index": True,
                    },
                    examples=examples,
                    tests=_generate_implausible_output_tests,
                    taxonomy=["avid-effect:performance:P0204"],
                    detector_name=self.__class__.__name__,
                )
            ]

        return []


def _generate_implausible_output_tests(issue: Issue):
    return {"Output plausibility": test_llm_output_plausibility(dataset=issue.dataset)}



from ...datasets.base import Dataset
from ...models.base import BaseModel


def get_dataset_subsample(dataset: Dataset, model: BaseModel, max_data_size: int):
    if max_data_size <= 0 or len(dataset) <= max_data_size:  # just double checking
        return dataset
    if not model.is_classification:
        return dataset.slice(lambda df: df.sample(max_data_size, random_state=42), row_level=False)

    # Slice the dataset while keeping the class proportions
    keep_ratio = max_data_size / len(dataset)
    dataset = dataset.slice(
        lambda df: df.groupby(dataset.target, group_keys=False, sort=False).sample(frac=keep_ratio, random_state=42),
        row_level=False,
    )
    return dataset



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

import pandas as pd

from giskard.datasets.base import Dataset
from giskard.llm.talk.config import ToolDescription
from giskard.llm.talk.tools.base import BaseTool, PredictionMixin

if TYPE_CHECKING:
    from giskard.models.base import BaseModel


class PredictTool(BaseTool, PredictionMixin):
    """Prediction calculation Tool.

    Attributes
    ----------
    default_name : str
        The default name of the Tool. Can be re-defined with constructor.
    default_description: str
        The default description of the Tool's functioning. Can be re-defined with constructor.
    """

    default_name: str = "predict"
    default_description: str = ToolDescription.PREDICT.value

    def __init__(
        self, model: BaseModel, dataset: Dataset, name: Optional[str] = None, description: Optional[str] = None
    ):
        """Constructor of the class.

        Parameters
        ----------
        model : BaseModel
            The Giskard Model.
        dataset : Dataset
            The Giskard Dataset.
        name : str, optional
            The name of the Tool.
            If not set, the `default_name` is used.
        description : str, optional
            The description of the Tool.
            If not set, the `default_description` is used.
        """
        BaseTool.__init__(self, name, description)
        PredictionMixin.__init__(self, model, dataset)

    @property
    def specification(self) -> str:
        """Return the Tool's specification in a JSON Schema format.

        Returns
        -------
        str
            The Tool's specification.
        """
        return {
            "type": "function",
            "function": {
                "name": self.name,
                "description": self.description,
                "parameters": {
                    "type": "object",
                    "properties": {
                        "features_dict": {
                            "type": "object",
                            "properties": {
                                feature: {"type": dtype} for feature, dtype in self.features_json_type.items()
                            },
                        }
                    },
                    "required": ["features_dict"],
                },
            },
        }

    def _get_input_from_user(self, feature_values: dict[str, any]) -> pd.DataFrame:
        """Form input based on the user's query.

        Take input feature values from user query. For the missed values, take the feature mean from the Dataset.

        Parameters
        ----------
        feature_values : dict[str, any]
            The dictionary with features and related values extracted from the query.

        Returns
        -------
        pd.DataFrame
            The DataFrame with the formed input.
        """
        # Import the private function from the 'model_explanation' module to avoid code duplication.
        # Also, if import from the top-level, we get the next error: "ImportError: cannot import name 'BaseModel'
        # from partially initialized module 'giskard.models.base' (most likely due to a circular import)"
        from giskard.models.model_explanation import _get_background_example

        # Prepare the background sample (the median and the mode of dataset's features).
        background_df = self._model.prepare_dataframe(
            self._dataset.df, self._dataset.column_dtypes, self._dataset.target
        )
        background_sample = _get_background_example(background_df, self._dataset.column_types)

        # Fill the background sample with known values (provided by the user query).
        for col_name, col_value in feature_values.items():
            background_sample[col_name] = col_value

        return background_sample

    def _prepare_input(self, feature_values: dict[str, any]) -> Dataset:
        """Prepare model input.

        Either extract rows form the dataset, or create input from the user's query.

        Parameters
        ----------
        feature_values : dict[str, any]
            The dictionary with features and related values extracted from the query.

        Returns
        -------
        Dataset
            The Giskard Dataset with model input.
        """
        final_input = self._get_input_from_dataset(feature_values)

        # If no data were selected from the dataset, build vector from the user input.
        if len(final_input) == 0:
            final_input = self._get_input_from_user(feature_values)

        return Dataset(final_input, target=None)

    def __call__(self, features_dict: dict[str, any]) -> str:
        """Execute the Tool's functionality.

        Predict an outcome based on rows from the dataset or on input formed from the user query.

        Parameters
        ----------
        features_dict : dict[str, any]
            The dictionary with features and related values extracted from the query.

        Returns
        -------
        str
            The model's prediction.
        """
        self._validate_features_dict(features_dict)
        model_input = self._prepare_input(features_dict)
        prediction = self._model.predict(model_input).prediction
        return ", ".join(map(str, prediction))



"""
This package provides LLM-specific detectors for the automatic model scan.
"""
from .llm_basic_sycophancy_detector import LLMBasicSycophancyDetector
from .llm_chars_injection_detector import LLMCharsInjectionDetector
from .llm_harmful_content_detector import LLMHarmfulContentDetector
from .llm_implausible_output_detector import LLMImplausibleOutputDetector
from .llm_information_disclosure_detector import LLMInformationDisclosureDetector
from .llm_output_formatting_detector import LLMOutputFormattingDetector
from .llm_prompt_injection_detector import LLMPromptInjectionDetector
from .llm_stereotypes_detector import LLMStereotypesDetector

__all__ = [
    "LLMBasicSycophancyDetector",
    "LLMCharsInjectionDetector",
    "LLMHarmfulContentDetector",
    "LLMImplausibleOutputDetector",
    "LLMInformationDisclosureDetector",
    "LLMOutputFormattingDetector",
    "LLMPromptInjectionDetector",
    "LLMStereotypesDetector",
]



import numpy as np
import pandas as pd

from .base_perturbation_function import PerturbationFunction


class NumericalTransformation(PerturbationFunction):
    def __init__(self, column: str, needs_dataset: bool = False) -> None:
        super().__init__(column, needs_dataset=needs_dataset)

    def execute(self, data: pd.DataFrame) -> pd.DataFrame:
        feature_data = data[self.column].dropna()
        data.loc[feature_data.index, self.column] = self.make_perturbation(feature_data)
        return data


class MultiplyByFactor(NumericalTransformation):
    name = "Multiply by factor"

    def __init__(self, column: str, factor: float) -> None:
        super().__init__(column)
        self.factor = factor

    def make_perturbation(self, values: pd.Series) -> pd.Series:
        # Round if the column is an integer type
        if np.issubdtype(values.dtype, np.integer):
            return np.round(values * self.factor).astype(values.dtype)
        return values * self.factor


class AddGaussianNoise(NumericalTransformation):
    name = "Add Gaussian noise"

    def __init__(self, column: str, mean: float = 0, std: float = 0.01, rng_seed: int = 1729) -> None:
        super().__init__(column)
        self.mean = mean
        self.std = std
        self.rng = np.random.default_rng(seed=rng_seed)

    def make_perturbation(self, values: pd.Series) -> pd.Series:
        noise = self.rng.normal(self.mean, self.std, values.shape)
        if np.issubdtype(values.dtype, np.integer):
            return np.round(values + noise).astype(values.dtype)
        return values + noise



from typing import Dict, Optional, Sequence

import logging
import uuid

import numpy as np
import pandas as pd
from sklearn.cluster import HDBSCAN

from ..llm.client import ChatMessage, LLMClient, get_default_client
from ..llm.embeddings import get_default_embedding
from ..llm.embeddings.base import BaseEmbedding
from ..llm.errors import LLMImportError
from ..utils.analytics_collector import analytics
from ..utils.language_detection import detect_lang
from .knowledge_base_plots import get_failure_plot, get_knowledge_plot

try:
    import umap
    from bokeh.io import output_notebook, reset_output, show
except ImportError as err:
    raise LLMImportError(flavor="llm") from err


logger = logging.getLogger("giskard.rag")

LANGDETECT_MAX_TEXT_LENGTH = 300
LANGDETECT_DOCUMENTS = 10

TOPIC_SUMMARIZATION_PROMPT = """Your task is to define the topic which best represents a set of documents.

Your are given below a list of documents and you must extract the topic best representing ALL contents.
- The topic name should be between 1 to 5 words
- Provide the topic in this language: {language}

Make sure to only return the topic name between quotes, and nothing else.

For example, given these documents:

<documents>
Camembert is a moist, soft, creamy, surface-ripened cow's milk cheese.
----------
Bleu d'Auvergne is a French blue cheese, named for its place of origin in the Auvergne region.
----------
Roquefort is a sheep milk cheese from the south of France.
</documents>

The topic is:
"French Cheese"

Now it's your turn. Here is the list of documents:

<documents>
{topics_elements}
</documents>

The topic is:
"""


class Document:
    """A class to wrap the elements of the knowledge base into a unified format."""

    def __init__(
        self, document: Dict[str, str], doc_id: str = None, features: Optional[Sequence] = None, topic_id: int = None
    ):
        features = features if features is not None else list(document.keys())

        self.content = (
            "\n".join(f"{k}: {v}" for k, v in document.items() if k in features)
            if len(features) > 1
            else document[features[0]]
        )

        self.metadata = document
        self.id = doc_id if doc_id is not None else str(uuid.uuid4())
        self.embeddings = None
        self.reduced_embeddings = None
        self.topic_id = topic_id


class KnowledgeBase:
    """
    A class to handle the knowledge base and the associated vector store.

    Parameters
    ----------
    knowledge_base_df: pd.DataFrame
        A dataframe containing the whole knowledge base.
    columns: Sequence[str], optional
        The list of columns from the `knowledge_base` to consider. If not specified, all columns of the knowledge base
        dataframe will be concatenated to produce a single document.
        Example: if your knowledge base consists in FAQ data with columns "Q" and "A", we will format each row into a
        single document "Q: [question]\\nA: [answer]" to generate questions.
    seed: int, optional
        The seed to use for random number generation.
    llm_client: LLMClient, optional:
        The LLM client to use for question generation. If not specified, a default openai client will be used.
    embedding_model: BaseEmbedding, optional
        The giskard embedding model to use for the knowledge base. By default we use giskard default model which is OpenAI "text-embedding-3-small".
    min_topic_size: int, optional
        The minimum number of document to form a topic inside the knowledge base.
    chunk_size: int = 2048
        The number of document to embed in a single batch.
    """

    def __init__(
        self,
        data: pd.DataFrame,
        columns: Optional[Sequence[str]] = None,
        seed: int = None,
        llm_client: Optional[LLMClient] = None,
        embedding_model: Optional[BaseEmbedding] = None,
        min_topic_size: Optional[int] = None,
        chunk_size: int = 2048,
    ) -> None:
        if len(data) == 0:
            raise ValueError("Cannot generate a Knowledge Base from empty DataFrame.")

        self._documents = [
            Document(
                knowledge_chunk,
                features=columns,
                doc_id=doc_id,
            )
            for doc_id, knowledge_chunk in data.to_dict("index").items()
        ]

        self._documents = [doc for doc in self._documents if doc.content.strip() != ""]

        if len(self) == 0:
            raise ValueError("Cannot generate a Knowledge Base with empty documents.")

        self._knowledge_base_df = data
        self._columns = columns

        self._rng = np.random.default_rng(seed=seed)
        self._llm_client = llm_client or get_default_client()
        self._embedding_model = embedding_model or get_default_embedding()

        # Estimate the minimum number of documents to form a topic
        self._min_topic_size = min_topic_size or round(2 + np.log(len(self._documents)))
        self.chunk_size = chunk_size

        self._embeddings_inst = None
        self._topics_inst = None
        self._index_inst = None
        self._reduced_embeddings_inst = None

        # Detect language of the documents, use only the first characters of a few documents to speed up the process
        document_languages = [
            detect_lang(doc.content[:LANGDETECT_MAX_TEXT_LENGTH])
            for doc in self._rng.choice(self._documents, size=LANGDETECT_DOCUMENTS)
        ]
        languages, occurences = np.unique(
            ["en" if (pd.isna(lang) or lang == "unknown") else lang for lang in document_languages], return_counts=True
        )
        self._language = languages[np.argmax(occurences)]

        self._documents_index = {doc.id: doc for doc in self._documents}

        analytics.track(
            "raget:knowledge-base-creation",
            {
                "num_documents": len(self._documents),
                "language": self._language,
            },
        )

    @classmethod
    def from_pandas(cls, df: pd.DataFrame, columns: Optional[Sequence[str]] = None, **kwargs) -> "KnowledgeBase":
        """Create a KnowledgeBase from a pandas DataFrame.

        Parameters
        ----------
        df: pd.DataFrame
            The DataFrame containing the knowledge base.
        columns: Sequence[str], optional
            The list of columns from the `knowledge_base` to consider. If not specified, all columns of the knowledge base
            dataframe will be concatenated to produce a single document.
            Example: if your knowledge base consists in FAQ data with columns "Q" and "A", we will format each row into a
            single document "Q: [question]\\nA: [answer]" to generate questions.
        kwargs:
            Additional settings for knowledge base (see __init__).
        """
        return cls(data=df, columns=columns, **kwargs)

    @property
    def _embeddings(self):
        if self._embeddings_inst is None:
            logger.debug("Computing Knowledge Base embeddings.")
            self._embeddings_inst = np.array(self._embedding_model.embed([doc.content for doc in self._documents]))
            for doc, emb in zip(self._documents, self._embeddings_inst):
                doc.embeddings = emb
        return self._embeddings_inst

    @property
    def _reduced_embeddings(self):
        if self._reduced_embeddings_inst is None:
            logger.debug("Calculating UMAP projection")
            reducer = umap.UMAP(
                n_neighbors=50,
                min_dist=0.5,
                n_components=2,
                random_state=1234,
                n_jobs=1,
            )
            self._reduced_embeddings_inst = reducer.fit_transform(self._embeddings)
            for doc, emb in zip(self._documents, self._reduced_embeddings_inst):
                doc.reduced_embeddings = emb
        return self._reduced_embeddings_inst

    @property
    def _dimension(self):
        return self._embeddings[0].shape[0]

    def get_savable_data(self):
        return {
            "columns": self._columns,
            "min_topic_size": self._min_topic_size,
            "topics": {int(k): topic for k, topic in self.topics.items()},
            "documents_topics": [int(doc.topic_id) for doc in self._documents],
        }

    @property
    def _index(self):
        if self._index_inst is None:
            try:
                from faiss import IndexFlatL2
            except ImportError as err:
                raise LLMImportError(flavor="llm") from err

            self._index_inst = IndexFlatL2(self._dimension)
            self._index_inst.add(self._embeddings)
        return self._index_inst

    @property
    def topics(self):
        if self._topics_inst is None:
            self._topics_inst = self._find_topics()
        return self._topics_inst

    def _find_topics(self):
        logger.info("Finding topics in the knowledge base.")
        hdbscan = HDBSCAN(
            min_cluster_size=self._min_topic_size,
            min_samples=3,
            metric="euclidean",
            cluster_selection_epsilon=0.0,
        )
        clustering = hdbscan.fit(self._reduced_embeddings)

        for i, doc in enumerate(self._documents):
            doc.topic_id = clustering.labels_[i]

        topics_ids = set(clustering.labels_)
        topics = {
            idx: self._get_topic_name([self._documents[doc_id] for doc_id in np.nonzero(clustering.labels_ == idx)[0]])
            for idx in topics_ids
            if idx != -1
        }
        topics[-1] = "Others"

        analytics.track(
            "raget:knowledge-topic-creation",
            {
                "num_topic": len(topics),
                "min_topic_size": self._min_topic_size,
            },
        )
        logger.info(f"Found {len(topics)} topics in the knowledge base.")
        return topics

    def _get_topic_name(self, topic_documents):
        logger.debug("Create topic name from topic documents")
        self._rng.shuffle(topic_documents)
        topics_str = "\n\n".join(["----------" + doc.content[:500] for doc in topic_documents[:10]])

        # prevent context window overflow
        topics_str = topics_str[: 3 * 8192]
        prompt = TOPIC_SUMMARIZATION_PROMPT.format(language=self._language, topics_elements=topics_str)

        raw_output = self._llm_client.complete([ChatMessage(role="user", content=prompt)], temperature=0.0).content

        return raw_output.strip().strip('"')

    def plot_topics(self, notebook: bool = True):
        if notebook:
            output_notebook()
        else:
            reset_output()
        show(get_knowledge_plot(self))

    def get_knowledge_plot(self):
        return get_knowledge_plot(self)

    def get_failure_plot(self, question_evaluation: Sequence[dict] = None):
        return get_failure_plot(self, question_evaluation)

    def get_random_document(self):
        return self._rng.choice(self._documents)

    def get_random_documents(self, n: int, with_replacement=False):
        if with_replacement:
            return list(self._rng.choice(self._documents, n, replace=True))

        docs = list(self._rng.choice(self._documents, min(n, len(self._documents)), replace=False))

        if len(docs) <= n:
            docs.extend(self._rng.choice(self._documents, n - len(docs), replace=True))

        return docs

    def get_neighbors(self, seed_document: Document, n_neighbors: int = 4, similarity_threshold: float = 0.2):
        seed_embedding = seed_document.embeddings

        relevant_documents = [
            doc
            for (doc, score) in self.vector_similarity_search_with_score(seed_embedding, k=n_neighbors)
            if score < similarity_threshold
        ]

        return relevant_documents

    def similarity_search_with_score(self, query: str, k: int) -> Sequence:
        query_emb = np.array(self._embedding_model.embed(query), dtype="float32")
        return self.vector_similarity_search_with_score(query_emb, k)

    def vector_similarity_search_with_score(self, query_emb: np.ndarray, k: int) -> Sequence:
        query_emb = np.atleast_2d(query_emb)
        distances, indices = self._index.search(query_emb, k)
        return [(self._documents[i], d) for d, i in zip(distances[0], indices[0])]

    def __len__(self):
        return len(self._documents)

    def __getitem__(self, doc_id: str):
        return self._documents_index[doc_id]



import uuid

from ..knowledge_base import KnowledgeBase
from ..testset import QuestionSample
from .base import GenerateFromSingleQuestionMixin, _LLMBasedQuestionGenerator
from .prompt import QAGenerationPrompt

QA_GENERATION_SYSTEM_PROMPT = """You are a powerful auditor, your role is to generate question & answer pair from a given list of context paragraphs.

The agent model you are auditing is the following:
- Agent description: {agent_description}

Your question must be related to a provided context.  
Please respect the following rules to generate the question:
- The answer to the question should be found inside the provided context
- The question must be self-contained
- The question and answer must be in this language: {language}

The user will provide the context, consisting in multiple paragraphs delimited by dashes "------".
You will return the question and the precise answer to the question based exclusively on the provided context.
You must output a single JSON object with keys 'question' and 'answer', without any other wrapping text or markdown. Make sure you only return valid JSON. """


QA_GENERATION_EXAMPLE_INPUT = """What payment methods do you accept?

We accept a variety of payment methods to provide our customers with a convenient and secure shopping experience. You can make a purchase using major credit and debit cards, including Visa, Mastercard, American Express, and Discover. We also offer the option to pay with popular digital wallets such as PayPal and Google Pay. For added flexibility, you can choose to complete your order using bank transfers or wire transfers. Rest assured that we prioritize the security of your personal information and go the extra mile to ensure your transactions are processed safely.
------
\tWhat is your shipping policy?

We offer free shipping on all orders over $50. For orders below $50, we charge a flat rate of $5.99. We offer shipping services to customers residing in all 50\n states of the US, in addition to providing delivery options to Canada and Mexico.
------
\tHow can I track my order?

Tracking your order is a breeze! Once your purchase has been successfully confirmed and shipped, you will receive a confirmation email containing your tracking number. You can simply click on the link provided in the email or visit our website's order tracking page. Enter your tracking number, and you will be able to monitor the progress of your shipment in real-time. This way, you can stay updated on the estimated delivery date and ensure you're available to receive your package.
"""

QA_GENERATION_EXAMPLE_OUTPUT = """{
    "question": "For which countries can I track my shipping?",
    "answer": "We ship to all 50 states in the US, as well as to Canada and Mexico. We offer tracking for all our shippings."
}"""


class SimpleQuestionsGenerator(GenerateFromSingleQuestionMixin, _LLMBasedQuestionGenerator):
    """
    Base question generator that generates questions from a KnowledgeBase.

    Parameters
    ----------
    context_neighbors: int, optional
        Number of context neighbors to use for question generation.
    context_similarity_threshold: float, optional
        Similarity threshold to keep neighboring document during question generation.
    context_window_length: int, optional
        Context window length of the llm used in the `llm_client` of the generator.
    llm_client: LLMClient, optional
        The LLM client to use for question generation. If not specified, a default openai client will be used.
    llm_temperature: float, optional
        The temperature to use in the LLM for question generation. The default is 0.5.
    """

    _prompt = QAGenerationPrompt(
        system_prompt=QA_GENERATION_SYSTEM_PROMPT,
        example_input=QA_GENERATION_EXAMPLE_INPUT,
        example_output=QA_GENERATION_EXAMPLE_OUTPUT,
    )

    _question_type = "simple"

    def generate_single_question(
        self,
        knowledge_base: KnowledgeBase,
        agent_description: str,
        language: str,
        seed_document=None,
    ) -> dict:
        """
        Generate a question from a list of context documents.

        Parameters
        ----------
        context_documents : Sequence[Document]
            The context documents to generate the question from.

        Returns
        -------
        QuestionSample
            The generated question and the metadata of the question.
        """
        seed_document = seed_document or knowledge_base.get_random_document()

        context_documents = knowledge_base.get_neighbors(
            seed_document, self._context_neighbors, self._context_similarity_threshold
        )
        context_str = "\n------\n".join(["", *[doc.content for doc in context_documents], ""])

        reference_context = "\n\n".join([f"Document {doc.id}: {doc.content}" for doc in context_documents])

        messages = self._prompt.to_messages(
            system_prompt_input={"agent_description": agent_description, "language": language},
            user_input=context_str,
        )

        generated_qa = self._llm_complete(messages=messages)
        question_metadata = {"question_type": self._question_type, "seed_document_id": seed_document.id}

        question = QuestionSample(
            id=str(uuid.uuid4()),
            question=generated_qa["question"],
            reference_answer=generated_qa["answer"],
            reference_context=reference_context,
            conversation_history=[],
            metadata=question_metadata,
        )
        return question


simple_questions = SimpleQuestionsGenerator()



import numpy as np
import pandas as pd


def get_df():
    df = pd.DataFrame({"x": np.arange(100), "y": np.arange(100)})
    return df


def get_model_and_df():
    from sklearn.linear_model import LinearRegression

    df = get_df()

    reg = LinearRegression()
    reg.fit(df["x"].to_numpy().reshape(100, 1), df["y"].to_numpy().reshape(100, 1))
    return reg, df


def get_pipeline():
    reg, _ = get_model_and_df()

    def preprocessor(df):
        return df["x"].to_numpy().reshape(len(df["x"]), 1)

    return preprocessor, reg



"""Statistical tests"""

from typing import Optional

import numbers

import numpy as np

from giskard.core.test_result import TestMessage, TestMessageLevel, TestResult
from giskard.datasets.base import Dataset
from giskard.models.base import BaseModel
from giskard.registry.decorators import test
from giskard.registry.slicing_function import SlicingFunction
from giskard.testing.utils.utils import check_slice_not_empty, validate_classification_label

from . import debug_description_prefix


@test(
    name="Right Label",
    tags=["heuristic", "classification"],
    debug_description=debug_description_prefix + "that <b>do not return the right classification label</b>.",
)
@validate_classification_label
def test_right_label(
    model: BaseModel,
    dataset: Dataset,
    classification_label: str,
    slicing_function: Optional[SlicingFunction] = None,
    threshold: float = 0.5,
) -> TestResult:
    """Test if the model returns the right classification label for a slice

    The test is passed when the percentage of rows returning the right
    classification label is higher than the threshold in a given slice

    Example: For a credit scoring model, the test is passed when more than 50%
    of people with high-salaries are classified as “non default”

    Parameters
    ----------
    model : BaseModel
        Model used to compute the test
    dataset : Dataset
        Dataset used to compute the test
    classification_label : str
        Classification label you want to test
    slicing_function : Optional[SlicingFunction]
        Slicing function to be applied on the dataset (Default value = None)
    threshold : float
        Threshold for the percentage of passed rows (Default value = 0.5)
    debug : bool
        If True and the test fails,
        a dataset will be provided containing the rows that do not return the right classification label. (Default value = False)

    Returns
    -------
    TestResult
        The test result.
    """
    if slicing_function:
        dataset = dataset.slice(slicing_function)
        check_slice_not_empty(sliced_dataset=dataset, dataset_name="dataset", test_name="test_right_label")

    prediction_results = model.predict(dataset).prediction

    passed_idx = dataset.df.loc[prediction_results == classification_label].index.values

    passed_ratio = len(passed_idx) / len(dataset)
    passed = bool(passed_ratio > threshold)

    # --- debug ---
    output_ds = list()
    if not passed:
        output_ds.append(dataset.slice(lambda df: df.loc[~dataset.df.index.isin(passed_idx)], row_level=False))
    # ---

    return TestResult(actual_slices_size=[len(dataset)], metric=passed_ratio, passed=passed, output_ds=output_ds)


@test(
    name="Output in range",
    tags=["heuristic", "classification", "regression"],
    debug_description=debug_description_prefix + "that are <b>out of the given range</b>.",
)
@validate_classification_label
def test_output_in_range(
    model: BaseModel,
    dataset: Dataset,
    slicing_function: Optional[SlicingFunction] = None,
    classification_label: Optional[str] = None,
    min_range: float = 0.3,
    max_range: float = 0.7,
    threshold: float = 0.5,
) -> TestResult:
    """Test if the model output belongs to the right range for a slice

    The test is passed when the ratio of rows in the right range inside the
    slice is higher than the threshold.

    For classification: Test if the predicted probability for a given classification label
    belongs to the right range for a dataset slice

    For regression: Test if the predicted output belongs to the right range for a dataset slice

    Example:
    For Classification: For a credit scoring model, the test is passed when more than 50% of
    people with high wage have a probability of defaulting between 0 and 0.1

    For Regression: The predicted Sale Price of a house in the city falls in a particular range

    Parameters
    ----------
    model : BaseModel
        Model used to compute the test
    dataset : Dataset
        Dataset used to compute the test
    slicing_function : Optional[SlicingFunction]
        Slicing function to be applied on the dataset (Default value = None)
    classification_label : Optional[str]
        Optional. Classification label you want to test (Default value = None)
    min_range : float
        Minimum probability of occurrence of classification label (Default value = 0.3)
    max_range : float
        Maximum probability of occurrence of classification label (Default value = 0.7)
    threshold : float
        Threshold for the percentage of passed rows (Default value = 0.5)
    debug : bool
        If True and the test fails,
        a dataset will be provided containing the rows that are out of the given range. (Default value = False)

    Returns
    -------
    TestResult
        The test result.

    """
    if slicing_function:
        dataset = dataset.slice(slicing_function)
        check_slice_not_empty(sliced_dataset=dataset, dataset_name="dataset", test_name="test_output_in_range")

    prediction_results = model.predict(dataset)

    if model.is_regression:
        output = prediction_results.raw_prediction

    elif model.is_classification:
        output = prediction_results.all_predictions[classification_label].values

    else:
        raise ValueError(f"Prediction task is not supported: {model.meta.model_type}")

    passed_idx = dataset.df.loc[(output <= max_range) & (output >= min_range)].index.values

    passed_ratio = len(passed_idx) / len(dataset)
    passed = bool(passed_ratio >= threshold)

    # --- debug ---
    output_ds = list()
    if not passed:
        output_ds.append(dataset.slice(lambda df: df.loc[~dataset.df.index.isin(passed_idx)], row_level=False))
    # ---

    return TestResult(actual_slices_size=[len(dataset)], metric=passed_ratio, passed=passed, output_ds=output_ds)


@test(
    name="Disparate impact",
    tags=["heuristic", "classification"],
    debug_description=debug_description_prefix + "that are <b>incorrectly predicted on the positive outcome "
    "from both the protected and unprotected slices</b>.",
)
def test_disparate_impact(
    model: BaseModel,
    dataset: Dataset,
    protected_slicing_function: SlicingFunction,
    unprotected_slicing_function: SlicingFunction,
    positive_outcome: str,
    slicing_function: Optional[SlicingFunction] = None,
    min_threshold: float = 0.8,
    max_threshold: float = 1.25,
) -> TestResult:
    """Tests for disparate impact on a protected data slice.

    Tests if the model is biased more towards an unprotected slice of the dataset over a protected slice.
    Note that this test reflects only a possible bias in the model while being agnostic to any bias in the dataset
    it trained on. The Disparate Impact (DI) is only valid for classification models and is computed as the ratio
    between the average count of correct predictions for the protected slice and the unprotected one given a
    certain positive_outcome.

    Calculate the Disparate Impact between a protected and unprotected slice of a dataset. Otherwise
    known as the "80 percent" rule, the Disparate Impact determines if a model was having an "adverse impact" on a
    protected (or minority in some cases) group.

    Example: The rule was originally based on the rates at which job applicants were hired. For example, if XYZ
    Company hired 50 percent of the men applying for work in a predominantly male occupation while hiring only 20
    percent of the female applicants, one could look at the ratio of those two hiring rates to judge whether there
    might be a discrimination problem. The ratio of 20:50 means that the rate of hiring for female applicants is
    only 40 percent of the rate of hiring for male applicants. That is, 20 divided by 50 equals
    0.40, which is equivalent to 40 percent. Clearly, 40 percent is well below the 80 percent that was arbitrarily
    set as an acceptable difference in hiring rates. Therefore, in this example, XYZ Company could have been called
    upon to prove that there was a legitimate reason for hiring men at a rate so much higher than the rate of hiring
    women.

    Parameters
    ----------
    model : BaseModel
        Model used to compute the test
    dataset : Dataset
        Dataset used to compute the test
    protected_slicing_function : SlicingFunction
        Slicing function that defines the protected group from the full dataset given
    unprotected_slicing_function : SlicingFunction
        Slicing function that defines the unprotected group from the full dataset given
    positive_outcome : str
        The target value that is considered a positive outcome in the dataset
    slicing_function : Optional[SlicingFunction]
        Slicing function to be applied on the dataset (Default value = None)
    min_threshold : float
        Threshold below which the DI test is considered to fail, by default 0.8
    max_threshold : float
        Threshold above which the DI test is considered to fail, by default 1.25
    Returns
    -------
    TestResult
        The test result.

    """
    if slicing_function:
        dataset = dataset.slice(slicing_function)
        check_slice_not_empty(sliced_dataset=dataset, dataset_name="dataset", test_name="test_disparate_impact")

    # Try to automatically cast `positive_outcome` to the right type
    if isinstance(model.meta.classification_labels[0], numbers.Number):
        try:
            positive_outcome = float(positive_outcome)
        except ValueError:
            pass

    if positive_outcome not in list(model.meta.classification_labels):
        raise ValueError(
            f"The positive outcome chosen {positive_outcome} is not part of the dataset target values {list(model.meta.classification_labels)}."
        )

    protected_ds = dataset.slice(protected_slicing_function)
    unprotected_ds = dataset.slice(unprotected_slicing_function)

    if protected_ds.df.equals(unprotected_ds.df):
        raise ValueError(
            "The protected and unprotected datasets are equal. Please check that you chose different slices."
        )

    _protected_predictions = model.predict(protected_ds).prediction
    _unprotected_predictions = model.predict(unprotected_ds).prediction

    protected_predictions = np.squeeze(_protected_predictions == positive_outcome)
    unprotected_predictions = np.squeeze(_unprotected_predictions == positive_outcome)

    protected_proba = np.count_nonzero(protected_predictions) / len(protected_ds.df)
    unprotected_proba = np.count_nonzero(unprotected_predictions) / len(unprotected_ds.df)
    disparate_impact_score = protected_proba / unprotected_proba

    messages = [
        TestMessage(
            type=TestMessageLevel.INFO, text=f"min_threshold = {min_threshold}, max_threshold = {max_threshold}"
        )
    ]

    passed = bool((disparate_impact_score > min_threshold) * (disparate_impact_score < max_threshold))

    # --- debug ---
    output_ds = list()
    if not passed:
        failed_protected = list(_protected_predictions != protected_ds.df[dataset.target])
        failed_unprotected = list(_unprotected_predictions != unprotected_ds.df[dataset.target])
        failed_idx_protected = [i for i, x in enumerate(failed_protected) if x]
        failed_idx_unprotected = [i for i, x in enumerate(failed_unprotected) if x]
        output_ds.append(
            dataset.slice(lambda df: df.iloc[failed_idx_protected + failed_idx_unprotected], row_level=False)
        )
    # ---

    return TestResult(metric=disparate_impact_score, passed=passed, messages=messages, output_ds=output_ds)


@test(
    name="Statistical parity difference",
    tags=["heuristic", "classification"],
    debug_description=debug_description_prefix + "that are <b>incorrectly predicted on the positive outcome "
    "from both the protected and unprotected slices</b>.",
)
def test_statistical_parity_difference(
    model: BaseModel,
    dataset: Dataset,
    protected_slicing_function: SlicingFunction,
    unprotected_slicing_function: SlicingFunction,
    positive_outcome: str,
    slicing_function: Optional[SlicingFunction] = None,
    threshold: float = 0.1,
) -> TestResult:
    """Tests for Statistical Parity Difference (SPD) on a protected data slice.

    Tests if the model is biased more towards an unprotected slice of the dataset over a protected slice.
    Note that this test reflects only a possible bias in the model while being agnostic to any bias in the dataset
    it trained on. The SPD is only valid for classification models and is computed as the absolute difference
    between the average count of correct predictions for the protected slice and the unprotected one given a
    certain positive_outcome.

    Parameters
    ----------
    model : BaseModel
        Model used to compute the test
    dataset : Dataset
        Dataset used to compute the test
    protected_slicing_function : SlicingFunction
        Slicing function that defines the protected group from the full dataset given
    unprotected_slicing_function : SlicingFunction
        Slicing function that defines the unprotected group from the full dataset given
    positive_outcome : str
        The target value that is considered a positive outcome in the dataset
    slicing_function : Optional[SlicingFunction]
        Slicing function to be applied on the dataset (Default value = None)
    threshold : float
        Threshold below which the SPD test is considered to fail, by default 0.1


    Returns
    -------
    TestResult
        The test result.

    """
    if slicing_function:
        dataset = dataset.slice(slicing_function)
        check_slice_not_empty(
            sliced_dataset=dataset, dataset_name="dataset", test_name="test_statistical_parity_difference"
        )

    # Try to automatically cast `positive_outcome` to the right type
    if isinstance(model.meta.classification_labels[0], numbers.Number):
        try:
            positive_outcome = float(positive_outcome)
        except ValueError:
            pass

    if positive_outcome not in list(model.meta.classification_labels):
        raise ValueError(
            f"The positive outcome chosen {positive_outcome} is not part of the dataset target values {list(model.meta.classification_labels)}."
        )

    protected_ds = dataset.slice(protected_slicing_function)
    unprotected_ds = dataset.slice(unprotected_slicing_function)

    if protected_ds.df.equals(unprotected_ds.df):
        raise ValueError(
            "The protected and unprotected datasets are equal. Please check that you chose different slices."
        )

    _protected_predictions = model.predict(protected_ds).prediction
    _unprotected_predictions = model.predict(unprotected_ds).prediction

    protected_predictions = np.squeeze(_protected_predictions == positive_outcome)
    unprotected_predictions = np.squeeze(_unprotected_predictions == positive_outcome)

    protected_proba = np.count_nonzero(protected_predictions) / len(protected_ds.df)
    unprotected_proba = np.count_nonzero(unprotected_predictions) / len(unprotected_ds.df)
    statistical_parity_difference = abs(protected_proba - unprotected_proba)

    messages = [TestMessage(type=TestMessageLevel.INFO, text=f"threshold = {threshold}")]

    passed = bool(statistical_parity_difference < threshold)

    # --- debug ---
    output_ds = list()
    if not passed:
        failed_protected = list(_protected_predictions != protected_ds.df[dataset.target])
        failed_unprotected = list(_unprotected_predictions != unprotected_ds.df[dataset.target])
        failed_idx_protected = [i for i, x in enumerate(failed_protected) if x]
        failed_idx_unprotected = [i for i, x in enumerate(failed_unprotected) if x]
        output_ds.append(
            dataset.slice(lambda df: df.iloc[failed_idx_protected + failed_idx_unprotected], row_level=False)
        )
    # ---

    return TestResult(metric=statistical_parity_difference, passed=passed, messages=messages, output_ds=output_ds)


def _cramer_v(x, y):
    import pandas as pd
    from scipy import stats

    ct = pd.crosstab(x, y)
    return stats.contingency.association(ct, method="cramer")


def _mutual_information(x, y):
    from sklearn.metrics import adjusted_mutual_info_score

    return adjusted_mutual_info_score(x, y)


def _theil_u(x, y):
    import pandas as pd
    from scipy import stats
    from sklearn.metrics import mutual_info_score

    return mutual_info_score(x, y) / stats.entropy(pd.Series(y).value_counts(normalize=True))


@test(
    name="Nominal Association",
    tags=["statistic", "nominal association", "classification"],
    debug_description=debug_description_prefix + "of <b>the data slice that was given as an input of the test</b>.",
)
def test_nominal_association(
    model: BaseModel,
    dataset: Dataset,
    slicing_function: SlicingFunction,
    method: Optional[str] = "theil_u",
    threshold: float = 0.5,
    debug: bool = False,  # noqa: NOSONAR - old version tests will call this under legacy debug mode
):
    """Tests for nominal association between a dataset slice and the model predictions.

    A statistical test for nominal association between the dataset slice and the model predictions. It aims to
    determine whether there is a significant relationship or dependency between the two. It assesses whether the
    observed association is likely to occur by chance or if it represents a true association.

    Description: The general procedure involves setting up a null hypothesis that assumes no association between the
    variables and an alternative hypothesis that suggests an association exists. The statistical test is calculated
    based on three methods: "theil_u", "cramer_v" and "mutual_information".

    Parameters
    ----------
    model : BaseModel
        Model used to compute the test
    dataset : Dataset
        Dataset used to compute the test
    slicing_function : SlicingFunction
        Slicing function to be applied on the dataset
    method : Optional[str]
        The association test statistic. Choose between "theil_u", "cramer_v", and "mutual_information".
        (default = "theil_u")
    threshold : float
        Threshold value for the Cramer's V score (Default value = 0.5)
    debug : bool
        If True and the test fails,
        a dataset will be provided containing the rows of the dataset slice. (Default value = False)

    Returns
    -------
    TestResult
        The test result.

    """
    import pandas as pd

    sliced_dataset = dataset.slice(slicing_function)
    check_slice_not_empty(sliced_dataset=sliced_dataset, dataset_name="dataset", test_name="test_nominal_association")

    dx = pd.DataFrame(
        {
            "slice": dataset.df.index.isin(sliced_dataset.df.index).astype(int),
            "prediction": model.predict(dataset).prediction,
        },
        index=dataset.df.index,
    )
    dx.dropna(inplace=True)

    if method == "theil_u":
        metric = _theil_u(dx.slice, dx.prediction)
    elif method == "cramer_v":
        metric = _cramer_v(dx.slice, dx.prediction)
    elif method == "mutual_information":
        metric = _mutual_information(dx.slice, dx.prediction)
    else:
        raise ValueError(
            "Invalid argument value: 'method' argument must " "be 'theil_u', 'cramer_v', or 'mutual_information'"
        )

    passed = metric < threshold

    # --- debug ---
    output_ds = list()
    if not passed:
        output_ds.append(sliced_dataset)
    # ---

    messages = [TestMessage(type=TestMessageLevel.INFO, text=f"metric = {metric}, threshold = {threshold}")]

    return TestResult(metric=metric, passed=passed, messages=messages, output_ds=output_ds)


@test(
    name="Cramer's V",
    tags=["statistic", "nominal association", "classification"],
    debug_description=debug_description_prefix + "of <b>the data slice that was given as an input of the test</b>.",
)
def test_cramer_v(
    model: BaseModel, dataset: Dataset, slicing_function: SlicingFunction, threshold: float = 0.5
) -> TestResult:
    r"""Tests for nominal association between a dataset slice and the model predictions using Cramer's V.

    Cramer's V is a statistical measure used to assess the strength and nature of association between two
    categorical variables. It is an extension of the chi-squared test for independence and takes into account the
    dimensions of the contingency table. Cramer's V ranges from 0 to 1, where 0 indicates no association and 1
    indicates a perfect association.

    Description: Cramer's V is particularly useful for analyzing nominal data and understanding the relationship between
    categorical variables. It's a normalized version of the chi-squared statistic that considers the dimensions of the
    contingency table. The formula adjusts for the number of observations and the number of categories in the
    variables to provide a more interpretable measure of association.
    Mathematically, the Cramer's V metric can be expressed as:

    .. math::

       V = \sqrt{\frac{\chi^2}{n \cdot \min(k-1, r-1)}}

    where: :math:`\chi^2` is the chi-squared statistic for the two variables. n is the total number of observations.
    :math:`k` is the
    number of categories in one variable. :math:`r` is the number of categories in the other variable.

    Parameters
    ----------
    model : BaseModel
        Model used to compute the test
    dataset : Dataset
        Dataset used to compute the test
    slicing_function : SlicingFunction
        Slicing function to be applied on the dataset
    threshold : float
        Threshold value for the Cramer's V score (Default value = 0.5)
    debug : bool
        If True and the test fails,
        a dataset will be provided containing the rows of the dataset slice. (Default value = False)

    Returns
    -------
    TestResult
        The test result.

    """
    return test_nominal_association(model, dataset, slicing_function, "cramer_v", threshold).execute()


@test(
    name="Mutual Information",
    tags=["statistic", "nominal association", "classification"],
    debug_description=debug_description_prefix + "of <b>the data slice that was given as an input of the test</b>.",
)
def test_mutual_information(
    model: BaseModel, dataset: Dataset, slicing_function: SlicingFunction, threshold: float = 0.5
) -> TestResult:
    r"""Tests for nominal association between a dataset slice and the model predictions using Mutual Information.

    The mutual information statistical test is a measure used to quantify the degree of association between two
    categorical variables. It assesses how much information about one variable can be gained from knowing the other
    variable's value. Mutual information is based on the concept of entropy and provides a way to determine the level
    of dependency or correlation between categorical variables.

    Description: Mutual information measures the reduction in uncertainty about one variable given knowledge of the
    other variable. It takes into account both individual and joint distributions of the variables and provides a value
    indicating how much information is shared between them. Higher mutual information values suggest stronger
    association, while lower values indicate weaker or no association.
    Mathematically, the mutual information metric can be expressed as:

    .. math::

       I(X;Y) = \sum_{x \in X} \sum_{y \in Y} p(x, y) \cdot \log ( \frac{p(x, y)}{p(x) \cdot p(y)})

    where: :math:`p(x,y)` is the joint probability mass function of variables :math:`X` and :math:`Y`. :math:`p(x)` and
    :math:`p(y)` are the marginal probability mass functions of variables :math:`X` and :math:`Y` respectively.

    Parameters
    ----------
    model : BaseModel
        Model used to compute the test
    dataset : Dataset
        Dataset used to compute the test
    slicing_function : SlicingFunction
        Slicing function to be applied on the dataset
    threshold : float
        Threshold value for the mutual information score (Default value = 0.5)
    debug : bool
        If True and the test fails,
        a dataset will be provided containing the rows of the dataset slice. (Default value = False)

    Returns
    -------
    TestResult
        The test result.

    """
    return test_nominal_association(model, dataset, slicing_function, "mutual_information", threshold).execute()


@test(
    name="Theil's U",
    tags=["statistic", "nominal association", "classification"],
    debug_description=debug_description_prefix + "of <b>the data slice that was given as an input of the test</b>.",
)
def test_theil_u(
    model: BaseModel, dataset: Dataset, slicing_function: SlicingFunction, threshold: float = 0.5
) -> TestResult:
    r"""Tests for nominal association between a dataset slice and the model predictions using Theil's U.

    Theil's U statistical test for nominal association is a measure used to assess the strength and direction
    of association between two categorical variables. It quantifies the inequality in the distribution of one variable
    relative to the distribution of the other variable, providing insights into the pattern of association between
    them. Theil's U ranges from 0 to 1, where 0 indicates no association, and 1 indicates a perfect association.

    Description: Theil's U for nominal association is commonly used to analyze the relationships between variables like
    ethnicity, gender, or occupation. It considers the proportions of one variable's categories within each category of
    the other variable. The calculation involves comparing the observed joint distribution of the two variables with
    what would be expected if there were no association.
    Mathematically, Theil's U for nominal association can be expressed as:

    .. math::

       U = \frac{H(x|y) - H(y|x)}{H(x)}

    where :math:`H(x|y)`, :math:`H(y|x)` are the conditional entropies of the two variables and :math:`H(x)` is the
    entropy of the first variable.

    Parameters
    ----------
    model : BaseModel
        Model used to compute the test
    dataset : Dataset
        Dataset used to compute the test
    slicing_function : SlicingFunction
        Slicing function to be applied on the dataset
    threshold : float
        Threshold value for the Theil's U score (Default value = 0.5)
    debug : bool
        If True and the test fails,
        a dataset will be provided containing the rows of the dataset slice. (Default value = False)

    Returns
    -------
    TestResult
        The test result.
    """
    return test_nominal_association(model, dataset, slicing_function, "theil_u", threshold).execute()



def format_number(value, n=3):
    if isinstance(value, float):
        if value > 1e6 or value < 1e-3:
            return f"{value:.{n}e}"
        else:
            return f"{value:.{3}f}"

    if isinstance(value, int) and abs(value) > 1e4:
        return f"{value:.{n}e}"

    return value


def truncate(text, max_length=240, ellipsis="…"):
    """Truncates a text to the given length, adding an ellipsis at the end if needed."""
    if len(text) > max_length:
        return text[: max_length - len(ellipsis)] + ellipsis
    return text



from typing import Optional

from dataclasses import dataclass

from ...llm.client import ChatMessage


@dataclass
class QAGenerationPrompt:
    system_prompt: str
    user_input_template: Optional[str] = None
    example_input: Optional[str] = None
    example_output: Optional[str] = None

    def _format_example_prompt(self, examples):
        if examples is not None:
            return examples

        examples = []
        if self.example_input is not None:
            examples.append(ChatMessage(role="user", content=self.example_input))
        if self.example_output is not None:
            examples.append(ChatMessage(role="assistant", content=self.example_output))
        return examples

    def to_messages(
        self,
        system_prompt_input,
        user_input,
        add_examples=True,
        examples=None,
    ):
        messages = [
            ChatMessage(
                role="system",
                content=self.system_prompt.format(**system_prompt_input),
            )
        ]
        if add_examples:
            messages.extend(self._format_example_prompt(examples))

        if self.user_input_template is None:
            messages.append(ChatMessage(role="user", content=user_input))
        else:
            messages.append(ChatMessage(role="user", content=self.user_input_template.format(**user_input)))

        return messages



from typing import Optional

import numpy as np
import pandas as pd

from giskard.core.core import SupportedModelTypes
from giskard.core.test_result import TestMessage, TestMessageLevel, TestResult
from giskard.datasets.base import Dataset
from giskard.llm import LLMImportError
from giskard.models.base import BaseModel
from giskard.models.utils import fix_seed
from giskard.registry.decorators import test
from giskard.registry.slicing_function import SlicingFunction
from giskard.registry.transformation_function import TransformationFunction
from giskard.testing.utils.stat_utils import (
    equivalence_t_test,
    equivalence_wilcoxon,
    paired_t_test,
    paired_wilcoxon,
)
from giskard.testing.utils.utils import (
    Direction,
    check_slice_not_empty,
    validate_classification_label,
)
from giskard.utils.logging_utils import timer

from . import debug_description_prefix


def _predict_numeric_result(model: BaseModel, ds: Dataset, output_proba=True, classification_label=None):
    if model.is_text_generation:
        return model.predict(ds).raw_prediction.flatten()
    if model.is_regression or not output_proba:
        return model.predict(ds).raw_prediction
    elif model.is_classification and classification_label is not None:
        return model.predict(ds).all_predictions[classification_label].values
    elif model.is_classification:
        return model.predict(ds).probabilities


def _prediction_ratio(prediction, perturbed_prediction):
    return abs(perturbed_prediction - prediction) / prediction if prediction != 0 else abs(perturbed_prediction)


@timer("Perturb and predict data")
def _perturb_and_predict(
    model: BaseModel,
    ds: Dataset,
    transformation_function: TransformationFunction,
    output_proba=True,
    classification_label=None,
):
    fix_seed()

    results_df = pd.DataFrame(
        {
            "prediction": _predict_numeric_result(model, ds, output_proba, classification_label),
        },
        index=ds.df.index,
    )

    perturbed_ds = ds.transform(transformation_function)
    results_df["perturbed_prediction"] = _predict_numeric_result(
        model, perturbed_ds, output_proba, classification_label
    )
    modified_idx = ds.df.compare(perturbed_ds.df).index

    return results_df.loc[modified_idx], len(modified_idx)


@timer("Compare and predict the data")
def _compare_prediction(results_df, prediction_task, direction, output_sensitivity=None):
    if direction == Direction.Invariant:
        if prediction_task == SupportedModelTypes.CLASSIFICATION:
            passed_idx = results_df.loc[results_df["prediction"] == results_df["perturbed_prediction"]].index.values

        elif prediction_task == SupportedModelTypes.TEXT_GENERATION:
            try:
                import evaluate

                scorer = evaluate.load("bertscore")
            except ImportError as err:
                raise LLMImportError() from err
            except FileNotFoundError as err:
                raise LLMImportError(
                    "Your version of evaluate does not support 'bertscore'. "
                    "Please use 'pip install -U evaluate' to upgrade it"
                ) from err

            score = scorer.compute(
                predictions=results_df["perturbed_prediction"].values,
                references=results_df["prediction"].values,
                model_type="distilbert-base-multilingual-cased",
                idf=True,
            )
            passed = np.array(score["f1"]) > 1 - output_sensitivity
            passed_idx = results_df.loc[passed].index.values

        elif prediction_task == SupportedModelTypes.REGRESSION:
            results_df["predict_difference_ratio"] = results_df.apply(
                lambda x: _prediction_ratio(x["prediction"], x["perturbed_prediction"]),
                axis=1,
            )
            output_sensitivity = 0.1 if not output_sensitivity else output_sensitivity
            passed_idx = results_df.loc[results_df["predict_difference_ratio"] < output_sensitivity].index.values
        else:
            raise ValueError(f"Invalid prediction task: {prediction_task}")

    elif direction == Direction.Increasing:
        passed_idx = results_df.loc[results_df["prediction"] < results_df["perturbed_prediction"]].index.values

    elif direction == Direction.Decreasing:
        passed_idx = results_df.loc[results_df["prediction"] > results_df["perturbed_prediction"]].index.values
    else:
        raise ValueError(f"Invalid direction: {direction}")

    failed_idx = results_df.loc[~results_df.index.isin(passed_idx)].index.values
    return passed_idx, failed_idx


def _compare_probabilities_t_test(result_df, direction, window_size=0.1, critical_quantile=0.05):
    if direction == Direction.Invariant:
        p_value = equivalence_t_test(
            result_df["prediction"],
            result_df["perturbed_prediction"],
            window_size=window_size,
            critical_quantile=critical_quantile,
        )[1]

    elif direction == Direction.Increasing:
        p_value = paired_t_test(
            result_df["prediction"],
            result_df["perturbed_prediction"],
            alternative="less",
            critical_quantile=critical_quantile,
        )[1]

    elif direction == Direction.Decreasing:
        p_value = paired_t_test(
            result_df["prediction"],
            result_df["perturbed_prediction"],
            alternative="greater",
            critical_quantile=critical_quantile,
        )[1]

    return p_value


def _compare_probabilities_wilcoxon(result_df, direction, window_size=0.2, critical_quantile=0.05):
    if direction == Direction.Invariant:
        p_value = equivalence_wilcoxon(
            result_df["prediction"],
            result_df["perturbed_prediction"],
            window_size=window_size,
            critical_quantile=critical_quantile,
        )[1]

    elif direction == Direction.Increasing:
        p_value = paired_wilcoxon(
            result_df["prediction"],
            result_df["perturbed_prediction"],
            alternative="less",
            critical_quantile=critical_quantile,
        )[1]

    elif direction == Direction.Decreasing:
        p_value = paired_wilcoxon(
            result_df["prediction"],
            result_df["perturbed_prediction"],
            alternative="greater",
            critical_quantile=critical_quantile,
        )[1]

    return p_value


def _test_metamorphic(
    model,
    direction: Direction,
    dataset: Dataset,
    transformation_function: TransformationFunction,
    threshold: float,
    classification_label=None,
    output_sensitivity=None,
    output_proba=True,
    debug: bool = False,  # noqa: NOSONAR - old version tests will call this under legacy debug mode
) -> TestResult:
    results_df, modified_rows_count = _perturb_and_predict(
        model, dataset, transformation_function, output_proba=output_proba, classification_label=classification_label
    )

    passed_idx, failed_idx = _compare_prediction(results_df, model.meta.model_type, direction, output_sensitivity)
    passed_ratio = len(passed_idx) / modified_rows_count if modified_rows_count != 0 else 1

    messages = [TestMessage(type=TestMessageLevel.INFO, text=f"{modified_rows_count} rows were perturbed")]

    passed = bool(passed_ratio > threshold)
    # --- debug ---
    output_ds = list()
    if not passed:
        output_ds.append(dataset.slice(lambda df: df.loc[failed_idx], row_level=False))
    # ---

    return TestResult(
        actual_slices_size=[len(dataset.df)],
        metric=passed_ratio,
        passed=passed,
        messages=messages,
        output_ds=output_ds,
    )


@test(
    name="Invariance (proportion)",
    debug_description=debug_description_prefix + "that are <b>non-invariant after perturbation</b>.",
)
def test_metamorphic_invariance(
    model: BaseModel,
    dataset: Dataset,
    transformation_function: TransformationFunction,
    slicing_function: Optional[SlicingFunction] = None,
    threshold: float = 0.5,
    output_sensitivity: float = None,
):
    """Tests if model prediction is invariant to perturbations

    For classification: Test if the predicted classification label remains the
    same after feature values perturbation.
    For regression: Check whether the predicted output remains the same at the
    ``output_sensibility`` level after feature values perturbation.

    The test is passed when the ratio of invariant rows is higher than the
    threshold.

    Example: The test is passed when, by switching gender from male to female,
    more than 50%(threshold 0.5) of males have unchanged outputs.

    Parameters
    ----------
    model : BaseModel
        Model used to compute the test
    dataset : Dataset
        Dataset used to compute the test
    transformation_function : TransformationFunction
        Function performing the perturbations to be applied on dataset.
    slicing_function : Optional[SlicingFunction]
        Slicing function to be applied on dataset.
    threshold : float
        The threshold value for the ratio of invariant rows.
    output_sensitivity : float
        For regression models. The threshold for ratio between the difference
        between perturbed prediction and actual prediction over the actual
        prediction for a regression model. We consider there is a prediction
        difference for regression if the ratio is above the output_sensitivity
        of 0.1.
    debug : bool
        If True and the test fails, a dataset will be provided containing the
        non-invariant rows.

    Returns
    -------
    TestResult
        A TestResult object containing the test result.
    """
    if slicing_function:
        dataset = dataset.slice(slicing_function)
        check_slice_not_empty(sliced_dataset=dataset, dataset_name="dataset", test_name="test_metamorphic_invariance")

    return _test_metamorphic(
        direction=Direction.Invariant,
        dataset=dataset,
        model=model,
        transformation_function=transformation_function,
        threshold=threshold,
        output_sensitivity=output_sensitivity,
        output_proba=False,
    )


@test(
    name="Increasing (proportion)",
    debug_description=debug_description_prefix + "that are <b>non-increasing after perturbation</b>.",
)
@validate_classification_label
def test_metamorphic_increasing(
    model: BaseModel,
    dataset: Dataset,
    transformation_function: TransformationFunction,
    slicing_function: Optional[SlicingFunction] = None,
    threshold: float = 0.5,
    classification_label: str = None,
):
    """Tests if the model prediction increases when the features are perturbed.

    For classification models, it tests if the model probability of a given
    ``classification_label`` is increasing after feature values perturbation.

    For regression models, it tests if the model prediction is increasing after
    feature values perturbation.

    The test is passed when the percentage of rows that are increasing is higher
    than the threshold.

    Example: For a credit scoring model, the test is passed when a decrease of
    wage by 10%, default probability is increasing for more than 50% of people
    in the dataset.

    Parameters
    ----------
    model : BaseModel
        Model used to compute the test
    dataset : Dataset
        Dataset used to compute the test
    transformation_function : TransformationFunction
        Function performing the perturbations to be applied on dataset.
    slicing_function : Optional[SlicingFunction]
        Slicing function to be applied on dataset (Default value = None)
    threshold : float
        The threshold value for the ratio of increasing rows. Default is 0.5.
    classification_label : str
        One specific label value from the target column (only for classification
        models).
    debug : bool
        If True and the test fails, a dataset will be provided containing the
        non-increasing rows.

    Returns
    -------
    TestResult
        A TestResult object containing the test result.
    """
    if slicing_function:
        dataset = dataset.slice(slicing_function)
        check_slice_not_empty(sliced_dataset=dataset, dataset_name="dataset", test_name="test_metamorphic_increasing")

    return _test_metamorphic(
        direction=Direction.Increasing,
        dataset=dataset,
        model=model,
        transformation_function=transformation_function,
        classification_label=classification_label,
        threshold=threshold,
    )


@test(
    name="Decreasing (proportion)",
    debug_description=debug_description_prefix + "that are <b>non-decreasing after perturbation</b>.",
)
@validate_classification_label
def test_metamorphic_decreasing(
    model: BaseModel,
    dataset: Dataset,
    transformation_function: TransformationFunction,
    slicing_function: Optional[SlicingFunction] = None,
    threshold: float = 0.5,
    classification_label: str = None,
):
    """Tests if the model prediction decreases when features are perturbed

    For classification models, it tests if the model probability of a given
    ``classification_label`` is decreasing after feature values perturbation.

    For regression models, it tests if the model prediction is decreasing after
    feature values perturbation.

    The test is passed when the percentage of rows that are decreasing is higher
    than the threshold.

    Example: For a credit scoring model, the test is passed when an increase of
    wage by 10%, default probability is decreasing for more than 50% of people
    in the dataset.

    Parameters
    ----------
    model : BaseModel
        Model used to compute the test
    dataset : Dataset
        Dataset used to compute the test
    transformation_function : TransformationFunction
        Function performing the perturbations to be applied on dataset.
    slicing_function : Optional[SlicingFunction]
        Slicing function to be applied on dataset
    threshold : float
        Threshold of the ratio of decreasing rows
    classification_label : str
        Optional. One specific label value from the target column
    debug : bool
        If True and the test fails, a dataset will be provided containing the
        non-decreasing rows.

    Returns
    -------
    TestResult
        A TestResult object containing the test result.
    """
    if slicing_function:
        dataset = dataset.slice(slicing_function)
        check_slice_not_empty(sliced_dataset=dataset, dataset_name="dataset", test_name="test_metamorphic_decreasing")

    return _test_metamorphic(
        direction=Direction.Decreasing,
        dataset=dataset,
        model=model,
        transformation_function=transformation_function,
        classification_label=classification_label,
        threshold=threshold,
    )


def _test_metamorphic_t_test(
    direction: Direction,
    model,
    dataset: Dataset,
    transformation_function: TransformationFunction,
    window_size: float,
    critical_quantile: float,
    classification_label=None,
    output_proba=True,
    debug: bool = False,  # noqa: NOSONAR - old version tests will call this under legacy debug mode
) -> TestResult:
    result_df, modified_rows_count = _perturb_and_predict(
        model, dataset, transformation_function, output_proba=output_proba, classification_label=classification_label
    )

    p_value = _compare_probabilities_t_test(result_df, direction, window_size, critical_quantile)

    messages = [TestMessage(type=TestMessageLevel.INFO, text=f"{modified_rows_count} rows were perturbed")]

    return _create_test_result(critical_quantile, dataset, direction, messages, model, p_value, result_df, debug)


def _create_test_result(
    critical_quantile,
    dataset,
    direction,
    messages,
    model,
    p_value,
    result_df,
    debug: bool = False,  # noqa: NOSONAR - old version tests will call this under legacy debug mode
):
    passed = bool(p_value < critical_quantile)
    # --- debug ---
    output_ds = list()
    if not passed:
        _, failed_idx = _compare_prediction(result_df, model.meta.model_type, direction, None)
        output_ds.append(dataset.slice(lambda df: df.loc[failed_idx], row_level=False))
    # ---
    return TestResult(
        actual_slices_size=[len(dataset.df)],
        metric=p_value,
        passed=passed,
        messages=messages,
        output_ds=output_ds,
    )


@test(
    name="Decreasing (t-test)",
    debug_description=debug_description_prefix + "that are <b>non-decreasing after perturbation</b>.",
)
@validate_classification_label
def test_metamorphic_decreasing_t_test(
    model: BaseModel,
    dataset: Dataset,
    transformation_function: TransformationFunction,
    slicing_function: Optional[SlicingFunction] = None,
    critical_quantile: float = 0.05,
    classification_label: str = None,
):
    """Tests if the model prediction decreases when the feature are perturbed

    Performs a t-test on two related samples. Sample A is constituted by the
    original predictions (probability of ``classification_label`` for
    classification models, or predicted value for regression models). Sample B
    is constituted by the predictions after perturbation of one or more of the
    features (by ``tranformation_function``).

    It performs a t-test to study if mean(B) < mean(A).

    The test is passed when the p-value of the t-test between (A) and (B) is
    below the critical quantile.

    Example: For a credit scoring model, the test is passed when an increase of
    wage by 10%, causes a statistically significant decrease of the default
    probability.

    Parameters
    ----------
    model : BaseModel
        Model used to compute the test
    dataset : Dataset
        Dataset used to compute the test
    transformation_function : TransformationFunction
        Function performing the perturbations to be applied on dataset.
    slicing_function : Optional[SlicingFunction]
        Slicing function to be applied on dataset
    critical_quantile : float
        Critical quantile above which the null hypothesis cannot be rejected
    classification_label : str
         (Default value = None)
    debug : bool
        If True and the test fails, a dataset will be provided containing the
        non-decreasing rows.

    Returns
    -------
    TestResult
        A TestResult object containing the test result.
    """
    if slicing_function:
        dataset = dataset.slice(slicing_function)
        check_slice_not_empty(
            sliced_dataset=dataset, dataset_name="dataset", test_name="test_metamorphic_decreasing_t_test"
        )

    return _test_metamorphic_t_test(
        direction=Direction.Decreasing,
        dataset=dataset,
        model=model,
        transformation_function=transformation_function,
        window_size=float("nan"),
        critical_quantile=critical_quantile,
        classification_label=classification_label,
    )


@test(
    name="Increasing (t-test)",
    debug_description=debug_description_prefix + "that are <b>non-increasing after perturbation</b>.",
)
@validate_classification_label
def test_metamorphic_increasing_t_test(
    model: BaseModel,
    dataset: Dataset,
    transformation_function: TransformationFunction,
    slicing_function: Optional[SlicingFunction] = None,
    critical_quantile: float = 0.05,
    classification_label: str = None,
):
    """Tests if the model prediction increases when feature values are perturbed

    Performs a t-test on two related samples. Sample A is constituted by the
    original predictions (probability of ``classification_label`` for
    classification models, or predicted value for regression models). Sample B
    is constituted by the predictions after perturbation of one or more of the
    features (by ``tranformation_function``).

    It performs a t-test to study if mean(B) > mean(A).

    The test is passed when the p-value of the t-test between (A) and (B) is
    below the critical quantile.

    Example: For a credit scoring model, the test is passed when a decrease of
    wage by 10%, causes a statistically significant increase of the default
    probability.

    Parameters
    ----------
    model : BaseModel
        Model used to compute the test
    dataset : Dataset
        Dataset used to compute the test
    transformation_function : TransformationFunction
        Function performing the perturbations to be applied on dataset.
    slicing_function : Optional[SlicingFunction]
        Slicing function to be applied on dataset (Default value = None)
    critical_quantile : float
        Critical quantile above which the null hypothesis cannot be rejected
    classification_label : str
        Only required for classification models.
    debug : bool
        If True and the test fails, a dataset will be provided containing the
        non-increasing rows.

    Returns
    -------
    actual_slices_size
        Length of dataset tested
    message
        Test result message
    metric
        The t-test in terms of p-value between unchanged rows over the perturbed rows
    passed
        TRUE if the p-value of the t-test between (A) and (B) is below the critical value

    """
    if slicing_function:
        dataset = dataset.slice(slicing_function)
        check_slice_not_empty(
            sliced_dataset=dataset, dataset_name="dataset", test_name="test_metamorphic_increasing_t_test"
        )

    return _test_metamorphic_t_test(
        direction=Direction.Increasing,
        dataset=dataset,
        model=model,
        transformation_function=transformation_function,
        window_size=float("nan"),
        critical_quantile=critical_quantile,
        classification_label=classification_label,
    )


@test(
    name="Invariance (t-test)",
    debug_description=debug_description_prefix + "that are <b>non-invariant after perturbation</b>.",
)
def test_metamorphic_invariance_t_test(
    model: BaseModel,
    dataset: Dataset,
    transformation_function: TransformationFunction,
    slicing_function: Optional[SlicingFunction] = None,
    window_size: float = 0.2,
    critical_quantile: float = 0.05,
) -> TestResult:
    """Tests if the model predictions are statistically invariant when the feature values are perturbed.

    Performs a t-test on two related samples. Sample A is constituted by the
    original predictions (probability of ``classification_label`` for
    classification models, or predicted value for regression models). Sample B
    is constituted by the predictions after perturbation of one or more of the
    features (by ``tranformation_function``).

    It performs a t-test to study if mean(A) is between mean(B) - window_size/2
    and mean(B) + window_size/2.

    The test is passed when the following tests pass:
      - the p-value of the t-test between (A) and (B) + window_size/2 is below the critical quantile
      - the p-value of the t-test between (B) - window_size/2 and (A) is below the critical quantile

    The test is passed when the p-value of the t-test between (A) and (B) is
    below the critical quantile.

    Example: The test is passed when, by switching gender from male to female,
    the probability distributions remains statistically invariant. In other
    words, the test is passed if the mean of the perturbed sample is
    statistically within a window determined by the user.

    Parameters
    ----------
    model : BaseModel
        Model used to compute the test
    dataset : Dataset
        Dataset used to compute the test
    transformation_function : TransformationFunction
        Function performing the perturbations to be applied on dataset.
    slicing_function : Optional[SlicingFunction]
        Slicing function to be applied on dataset (Default value = None)
    window_size : float
        Probability window in which the mean of the perturbed sample can be in
        order to pass the test
    critical_quantile : float
        Critical quantile above which the null hypothesis cannot be rejected
    debug : bool
        If True and the test fails, a dataset will be provided containing the
        non-invariant rows.

    Returns
    -------
    TestResult
        A TestResult object containing the test result.
    """
    if slicing_function:
        dataset = dataset.slice(slicing_function)
        check_slice_not_empty(
            sliced_dataset=dataset, dataset_name="dataset", test_name="test_metamorphic_invariance_t_test"
        )

    return _test_metamorphic_t_test(
        direction=Direction.Invariant,
        dataset=dataset,
        model=model,
        transformation_function=transformation_function,
        window_size=window_size,
        critical_quantile=critical_quantile,
        output_proba=False,
    )


def _test_metamorphic_wilcoxon(
    direction: Direction,
    model,
    dataset: Dataset,
    transformation_function: TransformationFunction,
    window_size: float,
    critical_quantile: float,
    classification_label=None,
    output_proba=True,
    debug: bool = False,  # noqa: NOSONAR - old version tests will call this under legacy debug mode
) -> TestResult:
    result_df, modified_rows_count = _perturb_and_predict(
        model, dataset, transformation_function, output_proba=output_proba, classification_label=classification_label
    )

    p_value = _compare_probabilities_wilcoxon(result_df, direction, window_size, critical_quantile)

    messages = [TestMessage(type=TestMessageLevel.INFO, text=f"{modified_rows_count} rows were perturbed")]

    return _create_test_result(critical_quantile, dataset, direction, messages, model, p_value, result_df, debug)


@test(
    name="Decreasing (Wilcoxon)",
    debug_description=debug_description_prefix + "that are <b>non-decreasing after perturbation</b>.",
)
@validate_classification_label
def test_metamorphic_decreasing_wilcoxon(
    model: BaseModel,
    dataset: Dataset,
    transformation_function: TransformationFunction,
    slicing_function: Optional[SlicingFunction] = None,
    critical_quantile: float = 0.05,
    classification_label: str = None,
):
    """Tests if the model prediction decreases when feature values are perturbed

    Performs the Wilcoxon signed-rank test on two related samples. Sample (A) is
    constituted by the original predictions (probability of
    ``classification_label`` for classification models, or predicted value for
    regression models). Sample B is constituted by the predictions after
    perturbation of one or more features by ``tranformation_function``.

    This test computes the decreasing test to study if mean(B) < mean(A)
    The test is passed when the p-value of the Wilcoxon signed-rank test
    between (A) and (B) is below the critical quantile.

    Example: For a credit scoring model, the test is passed when a decrease of
    wage by 10% causes a statistically significant probability decrease of the
    default probability.

    Parameters
    ----------
    model : BaseModel
        Model used to compute the test
    dataset : Dataset
        Dataset used to compute the test
    transformation_function : TransformationFunction
        Function performing the perturbations to be applied on dataset.
    slicing_function : Optional[SlicingFunction]
        Slicing function to be applied on dataset
    critical_quantile : float
        Critical quantile above which the null hypothesis cannot be rejected
    classification_label : str
        Only required for classification models.
    debug : bool
        If True and the test fails, a dataset will be provided containing the
        non-decreasing rows. (Default value = False)

    Returns
    -------
    TestResult
        A TestResult object containing the test result.
    """
    if slicing_function:
        dataset = dataset.slice(slicing_function)
        check_slice_not_empty(
            sliced_dataset=dataset, dataset_name="dataset", test_name="test_metamorphic_decreasing_wilcoxon"
        )

    return _test_metamorphic_wilcoxon(
        direction=Direction.Decreasing,
        dataset=dataset,
        model=model,
        transformation_function=transformation_function,
        classification_label=classification_label,
        window_size=float("nan"),
        critical_quantile=critical_quantile,
    )


@test(
    name="Increasing (Wilcoxon)",
    debug_description=debug_description_prefix + "that are <b>non-increasing after perturbation</b>.",
)
@validate_classification_label
def test_metamorphic_increasing_wilcoxon(
    model: BaseModel,
    dataset: Dataset,
    transformation_function: TransformationFunction,
    slicing_function: Optional[SlicingFunction] = None,
    critical_quantile: float = 0.05,
    classification_label: str = None,
):
    """Tests if the model prediction increases when feature values are perturbed

    Performs the Wilcoxon signed-rank test on two related samples. Sample (A) is
    constituted by the original predictions (probability of
    ``classification_label`` for classification models, or predicted value for
    regression models). Sample B is constituted by the predictions after
    perturbation of one or more features by ``tranformation_function``.

    This test computes the decreasing test to study if mean(B) > mean(A)
    The test is passed when the p-value of the Wilcoxon signed-rank test
    between (A) and (B) is below the critical quantile.

    Example: For a credit scoring model, the test is passed when a decrease of
    wage by 10% causes a statistically significant probability increase of the
    default probability.

    Parameters
    ----------
    model : BaseModel
        Model used to compute the test
    dataset : Dataset
        Dataset used to compute the test
    transformation_function : TransformationFunction
        Function performing the perturbations to be applied on dataset.
    slicing_function : Optional[SlicingFunction]
        Slicing function to be applied on dataset
    critical_quantile : float
        Critical quantile above which the null hypothesis cannot be rejected
    classification_label : str
        Only required for classification models.
    debug : bool
        If True and the test fails, a dataset will be provided containing the
        non-increasing rows.

    Returns
    -------
    TestResult
        A TestResult object containing the test result.
    """
    if slicing_function:
        dataset = dataset.slice(slicing_function)
        check_slice_not_empty(
            sliced_dataset=dataset, dataset_name="dataset", test_name="test_metamorphic_increasing_wilcoxon"
        )

    return _test_metamorphic_wilcoxon(
        direction=Direction.Increasing,
        dataset=dataset,
        model=model,
        transformation_function=transformation_function,
        classification_label=classification_label,
        window_size=float("nan"),
        critical_quantile=critical_quantile,
    )


@test(
    name="Invariance (Wilcoxon)",
    debug_description=debug_description_prefix + "that are <b>non-invariant after perturbation</b>.",
)
def test_metamorphic_invariance_wilcoxon(
    model: BaseModel,
    dataset: Dataset,
    transformation_function: TransformationFunction,
    slicing_function: Optional[SlicingFunction] = None,
    window_size: float = 0.2,
    critical_quantile: float = 0.05,
) -> TestResult:
    """Tests if the model predictions are statistically invariant when the feature values are perturbed

    Performs the Wilcoxon signed-rank test on two related samples. Sample (A) is
    constituted by the original predictions (probability of
    ``classification_label`` for classification models, or predicted value for
    regression models). Sample B is constituted by the predictions after
    perturbation of one or more features by ``tranformation_function``.

    This test computes the equivalence test to show that mean(B) - window_size/2 < mean(A) < mean(B) + window_size/2

    The test is passed when the following tests pass:
    - the p-value of the t-test between (A) and (B)+window_size/2 is below the critical quantile
    - the p-value of the t-test between (B)-window_size/2 and (A) is below the critical quantile

    Example: The test is passed when, by switching gender from male to female,
    the probability distributions remains statistically invariant. In other
    words, the test is passed if the mean of the perturbed sample is
    statistically within a window determined by the user.

    Parameters
    ----------
    model : BaseModel
        Model used to compute the test
    dataset : Dataset
        Dataset used to compute the test
    transformation_function : TransformationFunction
        Function performing the perturbations to be applied on dataset.
    slicing_function : Optional[SlicingFunction]
        Slicing function to be applied on dataset (Default value = None)
    window_size : float
        Probability window in which the mean of the perturbed sample can be in
    critical_quantile : float
        Critical quantile above which the null hypothesis cannot be rejected
    debug : bool
        If True and the test fails, a dataset will be provided containing the
        non-invariant rows. (Default value = False)

    Returns
    -------
    TestResult
        A TestResult object containing the test result.
    """
    if slicing_function:
        dataset = dataset.slice(slicing_function)
        check_slice_not_empty(
            sliced_dataset=dataset, dataset_name="dataset", test_name="test_metamorphic_invariance_wilcoxon"
        )

    return _test_metamorphic_wilcoxon(
        direction=Direction.Invariant,
        dataset=dataset,
        model=model,
        transformation_function=transformation_function,
        window_size=window_size,
        critical_quantile=critical_quantile,
        output_proba=False,
    )



from typing import Any

import pandas as pd

from ...core.core import DatasetProcessFunctionMeta
from ...registry.registry import get_object_uuid
from ...registry.transformation_function import TransformationFunction


class PerturbationFunction(TransformationFunction):
    name: str

    def __init__(self, column: str, needs_dataset: bool = False) -> None:
        super().__init__(None, row_level=False, cell_level=False, needs_dataset=needs_dataset)
        self.column = column
        self.meta = DatasetProcessFunctionMeta(type="TRANSFORMATION")
        self.meta.uuid = get_object_uuid(self)
        self.meta.code = self.name
        self.meta.name = self.name
        self.meta.display_name = self.name
        self.meta.tags = ["pickle", "scan"]
        self.meta.doc = self.meta.default_doc("Automatically generated transformation function")

    def __str__(self) -> str:
        return self.name

    def make_perturbation(self, data_or_series: Any) -> Any:
        raise NotImplementedError()

    def execute(self, data: pd.DataFrame) -> pd.DataFrame:
        raise NotImplementedError()



import logging

import pandas as pd
from mlflow import MlflowClient
from mlflow.models.evaluation import EvaluationResult, ModelEvaluator
from mlflow.tracking.artifact_utils import get_artifact_uri

from ...scanner.report import ScanReport
from ...utils import fullname
from ...utils.analytics_collector import analytics
from .evaluation_artifacts import (
    GiskardDatasetEvaluationArtifact,
    GiskardScanResultEvaluationArtifact,
    GiskardScanSummaryEvaluationArtifact,
)
from .giskard_evaluator_utils import setup_dataset, setup_model, setup_scan

logger = logging.getLogger(__name__)


class GiskardEvaluator(ModelEvaluator):
    def can_evaluate(self, model_type, **kwargs):
        return model_type in ["classifier", "regressor", "text", "question-answering"]

    def _wrap_dataset(self, dataset):
        if not isinstance(dataset.features_data, pd.DataFrame):
            raise ValueError("Only pd.DataFrame are currently supported by the giskard evaluator.")
        else:
            try:
                giskard_dataset = setup_dataset(dataset, self.evaluator_config)

                # log dataset
                local_path = giskard_dataset.to_mlflow(mlflow_client=self.client, mlflow_run_id=self.run_id)
                return giskard_dataset, local_path
            except Exception as e:
                analytics.track(
                    "mlflow_integration:wrapping_dataset:error",
                    {
                        "mlflow_run_id": self.run_id,
                        "error": str(e),
                    },
                )
                raise ValueError(
                    "An error occurred while wrapping the dataset. "
                    "Please submit the traceback as a GitHub issue in the following "
                    "repository for further assistance: https://github.com/Giskard-AI/giskard."
                ) from e

    def _wrap_model(self, model, model_type, feature_names):
        try:
            giskard_model = setup_model(model, model_type, feature_names, self.evaluator_config)
            return giskard_model
        except Exception as e:
            analytics.track(
                "mlflow_integration:wrapping_model:error",
                {
                    "mlflow_run_id": self.run_id,
                    "error": str(e),
                },
            )
            raise ValueError(
                "An error occurred while wrapping the model. "
                "Please submit the traceback as a GitHub issue in the following "
                "repository for further assistance: https://github.com/Giskard-AI/giskard."
            ) from e

    def _perform_scan(self, giskard_model, giskard_dataset):
        try:
            scan_results = setup_scan(giskard_model, giskard_dataset, self.evaluator_config)

            # log html scan result
            scan_artifact_names = scan_results.to_mlflow(mlflow_client=self.client, mlflow_run_id=self.run_id)
            return scan_results, scan_artifact_names
        except Exception as e:
            analytics.track(
                "mlflow_integration:scan:error",
                {
                    "mlflow_run_id": self.run_id,
                    "error": str(e),
                },
            )
            raise ValueError(
                "An error occurred while scanning the model for vulnerabilities. "
                "Please submit the traceback as a GitHub issue in the following "
                "repository for further assistance: https://github.com/Giskard-AI/giskard."
            ) from e

    def _generate_metrics(self, scan_results):
        try:
            test_suite = scan_results.generate_test_suite("scan test suite")
            test_suite_results = test_suite.run()

            # log metrics resulting from scan
            return test_suite_results.to_mlflow(mlflow_client=self.client, mlflow_run_id=self.run_id)
        except Exception as e:
            analytics.track(
                "mlflow_integration:test_suite_generation:error",
                {
                    "mlflow_run_id": self.run_id,
                    "error": str(e),
                },
            )
            logger.warning(
                "An error occurred while generating the test suite. "
                "Visualising the results of the scan is still possible in the mlflow ui, but not as metrics. "
                "Please submit the traceback as a GitHub issue in the following "
                "repository for further assistance: https://github.com/Giskard-AI/giskard."
            )
            return {}

    def evaluate(
        self, *, model, model_type, dataset, run_id, evaluator_config, baseline_model=None, **kwargs
    ) -> EvaluationResult:
        self.client = MlflowClient()
        self.run_id = run_id
        self.evaluator_config = evaluator_config

        artifacts = dict()
        properties = dict()
        properties.update({"mlflow_run_id": str(self.run_id)} if self.run_id is not None else "none")

        # Wrapping dataset
        giskard_dataset, dataset_artifact_name = self._wrap_dataset(dataset)
        artifacts["giskard_dataset"] = GiskardDatasetEvaluationArtifact(
            uri=get_artifact_uri(self.run_id, dataset_artifact_name), content=giskard_dataset.df
        )
        properties.update({"dataset_id": str(giskard_dataset.id)} if giskard_dataset is not None else "none")

        # Wrapping model
        giskard_model = self._wrap_model(model, model_type=model_type, feature_names=dataset.feature_names)
        properties.update({"model_class": fullname(giskard_model), "model_inner_class": fullname(giskard_model.model)})
        analytics.track("mlflow_integration", properties)

        # Perform scan
        scan_results, scan_artifact_names = self._perform_scan(giskard_model, giskard_dataset)
        scan_results_artifact_name, scan_summary_artifact_name = scan_artifact_names[0], scan_artifact_names[1]
        scan_summary = ScanReport.get_scan_summary_for_mlflow(scan_results)
        artifacts["giskard_scan_results"] = GiskardScanResultEvaluationArtifact(
            uri=get_artifact_uri(self.run_id, scan_results_artifact_name), content=scan_results
        )
        if scan_summary_artifact_name:
            artifacts["giskard_scan_summary"] = GiskardScanSummaryEvaluationArtifact(
                uri=get_artifact_uri(self.run_id, scan_summary_artifact_name), content=scan_summary
            )

        # Generate test suite
        metrics = self._generate_metrics(scan_results)

        print(
            "The evaluation with giskard ran successfully! You can now visualise the results by running 'mlflow ui' "
            "in the terminal."
        )

        return EvaluationResult(metrics=metrics, artifacts=artifacts)



from typing import Optional, Sequence, Union

import json
from html import escape
from pathlib import Path

import matplotlib
import numpy as np
import pandas as pd

from ..llm.client.base import LLMClient
from ..llm.embeddings.base import BaseEmbedding
from ..llm.errors import LLMImportError
from ..visualization.widget import get_template
from .base import AgentAnswer
from .knowledge_base import KnowledgeBase
from .question_generators import COMPONENT_DESCRIPTIONS, QUESTION_ATTRIBUTION, RAGComponents
from .testset import QATestset, QuestionSample

try:
    from bokeh.document import Document
    from bokeh.embed import components
    from bokeh.io import curdoc, output_notebook, reset_output
    from bokeh.models import ColumnDataSource, Span, TabPanel, Tabs
    from bokeh.plotting import figure
except ImportError as err:
    raise LLMImportError(flavor="llm") from err


def get_colors(values, cmap_name="RdYlGn"):
    cmap = matplotlib.colormaps[cmap_name]
    normalizer = matplotlib.colors.Normalize()
    colors = ["#%02x%02x%02x" % (int(r), int(g), int(b)) for r, g, b, _ in 255 * cmap(normalizer(values))]
    return colors


class RAGReport:
    """
    Report class for the RAG model evaluation.

    Parameters
    ----------
    testset : QATestset
        The testset used to evaluate the agent.
    model_outputs : Sequence[ModelOutput]
        The evaluation results of the agent's answers. Should be a list of dictionaries with the following keys: "evaluation", "reason", "agent_answer".
    metrics_results : dict, optional
        The additional metrics computed during the evaluation. If provided, these metrics will be included in the report. The dict should have the following structure: `{"question_id": {"correctness": bool, "correctness_reason": str, "additional_metric": value, ...}}`.
    knowledge_base : KnowledgeBase
        The knowledge base used to create the testset.
    """

    def __init__(
        self,
        testset: QATestset,
        model_outputs: Sequence[AgentAnswer],
        metrics_results: dict,
        knowledge_base: Optional[KnowledgeBase] = None,
    ):
        self._testset = testset
        self._model_outputs = model_outputs
        self._metrics_results = metrics_results
        self._knowledge_base = knowledge_base

        self._recommendation = "Placeholder for the recommmendation."

        self._dataframe = testset.to_pandas().copy()
        self._dataframe["agent_answer"] = [answer.message for answer in model_outputs]

        metric_df = pd.DataFrame.from_dict(metrics_results, orient="index")
        self.metric_names = [
            metric for metric in metric_df.columns if metric not in ["correctness", "correctness_reason"]
        ]
        self._dataframe = self._dataframe.join(metric_df, on="id")

    def to_pandas(self):
        return self._dataframe

    def _repr_html_(self, notebook=True):
        if notebook:
            output_notebook()
        else:
            reset_output()

        return self.to_html()

    def to_html(self, filename=None, embed=False):
        """Renders the evaluation report as HTML.

        Saves or returns the HTML representation of the scan report.

        Parameters
        ----------
        filename : Optional[str]
            If provided, the HTML will be written to the file.
        """
        tpl = get_template("rag_report/rag_report.html")

        kb_script, kb_div = (
            components(self._apply_theme(self._get_knowledge_plot())) if self._knowledge_base else (None, None)
        )
        q_type_script, q_type_div = components(
            self._apply_theme(self.plot_correctness_by_metadata("question_type")), theme="dark_minimal"
        )
        topic_script, topic_div = components(
            self._apply_theme(self.plot_correctness_by_metadata("topic")), theme="dark_minimal"
        )

        metric_histograms = self.get_metrics_histograms()

        component_dict = self.component_scores().to_dict()["score"]

        for name, description in COMPONENT_DESCRIPTIONS.items():
            if name in component_dict:
                component_dict[name] = {"score": component_dict[name], "description": description}

        html = tpl.render(
            knowledge_script=kb_script,
            knowledge_div=kb_div,
            recommendation=self._recommendation,
            components=component_dict,
            correctness=self.correctness,
            q_type_correctness_script=q_type_script,
            q_type_correctness_div=q_type_div,
            topic_correctness_script=topic_script,
            topic_correctness_div=topic_div,
            additional_metrics=len(self.metric_names) > 0,
            metric_histograms=metric_histograms,
        )

        if filename is not None:
            with open(filename, "w", encoding="utf-8") as f:
                f.write(html)
            return

        if embed:
            from IPython.core.display import HTML

            return HTML(f'<iframe srcdoc="{escape(html)}" width=100% height=800px></iframe>')
        return html

    def save(self, folder_path: str):
        """Save all the report data to a folder.

        This includes the HTML report, the testset, the knowledge base, the evaluation results and the metrics if
        provided.

        Parameters
        ----------
        folder_path : str or Path
            The folder path to save the report data.
        """
        path = Path(folder_path)
        path.mkdir(exist_ok=True, parents=True)
        self.to_html(path / "report.html")
        self._testset.save(path / "testset.jsonl")

        report_details = {"recommendation": self._recommendation}
        with open(path / "report_details.json", "w", encoding="utf-8") as f:
            json.dump(report_details, f, ensure_ascii=False)

        with open(path / "agent_answer.json", "w", encoding="utf-8") as f:
            json.dump(
                [{"message": output.message, "documents": output.documents} for output in self._model_outputs],
                f,
                ensure_ascii=False,
            )

        if self._knowledge_base is not None:
            self._knowledge_base._knowledge_base_df.to_json(
                path / "knowledge_base.jsonl", orient="records", lines=True, force_ascii=False
            )
            with open(path / "knowledge_base_meta.json", "w", encoding="utf-8") as f:
                json.dump(self._knowledge_base.get_savable_data(), f, ensure_ascii=False)

        if self._metrics_results is not None:
            with open(path / "metrics_results.json", "w", encoding="utf-8") as f:
                json.dump(self._metrics_results, f, ensure_ascii=False)

    @classmethod
    def load(
        cls, folder_path: str, llm_client: Optional[LLMClient] = None, embedding_model: Optional[BaseEmbedding] = None
    ):
        """Load a saved report.

        It reconstructs the objects inside the report including the testset and the knowledge base.

        Parameters
        ----------
        folder_path : str or Path
            The folder path to load the report data from.
        llm_client : LLMClient, optional
            The LLMClient to use inside the knowledge base. If not provided, the default client will be used.
        embedding_model : BaseEmbedding, optional
            The embedding model to use inside the knowledge base. If not provided, the default model will be used.
        """
        path = Path(folder_path)

        testset = QATestset.load(path / "testset.jsonl")

        answers = json.load(open(path / "agent_answer.json", "r", encoding="utf-8"))
        model_outputs = [AgentAnswer(**answer) for answer in answers]

        report_details = json.load(open(path / "report_details.json", "r", encoding="utf-8"))
        testset._dataframe.index = testset._dataframe.index.astype(str)

        knowledge_base = None
        if (path / "knowledge_base.jsonl").exists():
            knowledge_base_meta = json.load(open(path / "knowledge_base_meta.json", "r", encoding="utf-8"))
            knowledge_base_data = pd.read_json(path / "knowledge_base.jsonl", orient="records", lines=True)

            topics = {int(k): topic for k, topic in knowledge_base_meta.pop("topics", None).items()}
            documents_topics = [int(topic_id) for topic_id in knowledge_base_meta.pop("documents_topics", None)]

            knowledge_base = KnowledgeBase(
                knowledge_base_data,
                llm_client=llm_client,
                embedding_model=embedding_model,
                columns=knowledge_base_meta.pop("columns", None),
                min_topic_size=knowledge_base_meta.pop("min_topic_size", None),
                chunk_size=knowledge_base_meta.pop("chunk_size", None),
                seed=knowledge_base_meta.pop("seed", None),
            )

            knowledge_base._topics_inst = topics

            for doc_idx, doc in enumerate(knowledge_base._documents):
                doc.topic_id = documents_topics[doc_idx]

        metrics_results = {}
        if (path / "metrics_results.json").exists():
            metrics_results = json.load(open(path / "metrics_results.json", "r", encoding="utf-8"))

        report = cls(testset, model_outputs, metrics_results, knowledge_base)
        report._recommendation = report_details["recommendation"]
        return report

    @property
    def topics(self):
        return self._testset.get_metadata_values("topic")

    @property
    def failures(self) -> pd.DataFrame:
        return self._dataframe[~self._dataframe["correctness"]]

    def get_failures(
        self,
        topic: Optional[Union[str, Sequence[str]]] = None,
        question_type: Optional[Union[str, Sequence[str]]] = None,
    ) -> pd.DataFrame:
        """
        Retrieves the failures from the results, optionally filtering by topic and question type.

        Parameters
        ----------
        topic : str or Sequence[str], optional
            The topic(s) to filter the failures by.
        question_type : str or Sequence[str], optional
            The question type(s) to filter the failures by.
        """
        failures = self.failures

        if topic:
            topic = [topic] if not isinstance(topic, Sequence) else topic
            failures = failures[failures["metadata"].apply(lambda x: x.get("topic") in topic)]
        if question_type:
            question_type = [question_type] if not isinstance(question_type, Sequence) else question_type
            failures = failures[failures["metadata"].apply(lambda x: x.get("question_type") in question_type)]

        return failures

    @property
    def correctness(self) -> float:
        """
        Compute the overall correctness of the agent's answers.
        """
        return self._dataframe["correctness"].mean()

    def correctness_by_question_type(self) -> pd.DataFrame:
        """
        Compute the correctness by question type.
        """
        correctness = self._correctness_by_metadata("question_type")
        return correctness

    def correctness_by_topic(self) -> pd.DataFrame:
        """
        Compute the correctness by topic.
        """
        return self._correctness_by_metadata("topic")

    def component_scores(self) -> pd.DataFrame:
        """
        Compute the scores for each RAG component.
        """
        correctness = self.correctness_by_question_type()
        available_question_types = {
            component: list(set([a for a in attribution]).intersection(correctness.index))
            for component, attribution in QUESTION_ATTRIBUTION.items()
        }

        scores = {
            component: (
                [sum(1 / len(attribution) * correctness.loc[q_type, "correctness"] for q_type in attribution)]
                if len(attribution) > 0
                else [1]
            )
            for component, attribution in available_question_types.items()
        }
        scores[RAGComponents.KNOWLEDGE_BASE] = [self.knowledge_base_score]

        score_df = pd.DataFrame.from_dict(scores, orient="index")
        score_df.columns = ["score"]
        score_df.index.rename("RAG Components", inplace=True)
        score_df.index = score_df.index.map(lambda x: RAGComponents(x).name)
        return score_df

    @property
    def knowledge_base_score(self):
        correctness_by_topic = [topic_score for _, topic_score in self.correctness_by_topic().itertuples()]
        return 1 - (max(correctness_by_topic) - min(correctness_by_topic))

    def _correctness_by_metadata(self, metadata_name: str):
        """
        Compute the correctness by a metadata field.
        """
        correctness = (
            self._dataframe["correctness"]
            .groupby(lambda idx: self._dataframe.loc[idx, "metadata"][metadata_name])
            .mean()
            .to_frame()
        )

        correctness.index.rename(metadata_name, inplace=True)
        return correctness

    def _get_knowledge_plot(self):
        tabs = [
            TabPanel(child=self._knowledge_base.get_knowledge_plot(), title="Topic exploration"),
            TabPanel(
                child=self._knowledge_base.get_failure_plot(
                    [
                        QuestionSample(**question, id="", reference_context="", conversation_history=[])
                        for question in self._dataframe[
                            ["question", "reference_answer", "agent_answer", "correctness", "metadata"]
                        ].to_dict(orient="records")
                    ]
                ),
                title="Failures",
            ),
        ]

        tabs = Tabs(tabs=tabs, sizing_mode="stretch_width", tabs_location="below")
        return tabs

    def plot_correctness_by_metadata(self, metadata_name: str):
        """
        Create a bokeh plot showing the correctness by a metadata field.
        """
        data = self._correctness_by_metadata(metadata_name)
        metadata_values = data.index.tolist()
        overall_correctness = self.correctness
        correctness = data["correctness"].to_numpy()

        source = ColumnDataSource(
            data={
                "correctness": correctness * 100,
                "metadata_values": metadata_values,
                "colors": get_colors(correctness),
            }
        )

        p = figure(
            y_range=metadata_values,
            height=350,
            toolbar_location=None,
            tools="hover",
            width_policy="max",
        )
        p.hbar(y="metadata_values", right="correctness", source=source, height=0.85, fill_color="#14191B")
        p.hbar(
            y="metadata_values",
            right="correctness",
            source=source,
            height=0.85,
            fill_color="#78BBFA",
            fill_alpha=0.7,
            line_color="white",
            line_width=2,
        )
        vline = Span(
            location=overall_correctness * 100,
            dimension="height",
            line_color="#EA3829",
            line_width=2,
            line_dash="dashed",
        )

        p.add_layout(vline)
        p.background_fill_color = "#14191B"

        p.x_range.start = 0
        r_line = p.line(
            [0],
            [0],
            legend_label="Correctness on the entire Testset",
            line_dash="dashed",
            line_color="#EA3829",
            line_width=2,
        )
        r_line.visible = False  # Set this fake line to invisible
        p.legend.background_fill_color = "#111516"
        p.legend.background_fill_alpha = 0.5

        p.xaxis.axis_label = "Correctness (%)"
        p.title.text_font_size = "14pt"
        p.hover.tooltips = [
            (metadata_name, "@metadata_values"),
            ("Correctness", "@correctness{0.00}"),
        ]

        return p

    def plot_metrics_hist(self, metric_name: str, filter_metadata: dict = None):
        """
        Create a bokeh histogram plot for a RAGAS metric.

        Parameters
        ----------
        metric_name : str
            The name of the RAGAS metric to plot.
        filter_metadata : dict, optional
            Aggregate the question that have the specified metadata values. The keys of the dictionary should be the metadata names and the values should be the metadata values to filter by.
        """
        if metric_name in self.metric_names:
            if filter_metadata is not None:
                data = self._dataframe[metric_name][
                    self._dataframe["metadata"].apply(lambda x: all(x.get(k) in v for k, v in filter_metadata.items()))
                ]
            else:
                data = self._dataframe[metric_name]

            p = figure(
                width=300,
                height=200,
                toolbar_location=None,
                title=metric_name.replace("_", " "),
                tools="hover",
                width_policy="max",
            )

            bins = np.linspace(0, 1, 21)
            hist, edges = np.histogram(data, bins=bins)
            p.quad(
                top=hist,
                bottom=0,
                left=edges[:-1],
                right=edges[1:],
                fill_color="#78bbfa",
                line_color="white",
            )
            p.title.text_font_size = "12pt"
            p.hover.tooltips = [
                ("Range", "@left{0.00} to @right{0.00}"),
                ("# questions", "@top"),
            ]

            return p

    def _apply_theme(self, p):
        curdoc().theme = "dark_minimal"
        doc = Document(theme=curdoc().theme)
        doc.add_root(p)
        return p

    def _get_plot_components(self, p):
        script, div = components(self._apply_theme(p), theme="dark_minimal")
        return {"script": script, "div": div}

    def get_metrics_histograms(self):
        histograms_dict = {}
        histograms_dict["Overall"] = {
            "Overall": {
                metric: self._get_plot_components(self.plot_metrics_hist(metric)) for metric in self.metric_names
            }
        }
        histograms_dict["Topics"] = {
            topic: {
                metric: self._get_plot_components(self.plot_metrics_hist(metric, {"topic": [topic]}))
                for metric in self.metric_names
            }
            for topic in self._testset.get_metadata_values("topic")
        }
        histograms_dict["Question"] = {
            q_type: {
                metric: self._get_plot_components(self.plot_metrics_hist(metric, {"question_type": [q_type]}))
                for metric in self.metric_names
            }
            for q_type in self._testset.get_metadata_values("question_type")
        }
        return histograms_dict



from .coherency import CoherencyEvaluator
from .correctness import CorrectnessEvaluator
from .plausibility import PlausibilityEvaluator
from .requirements import RequirementEvaluator

__all__ = [
    "CoherencyEvaluator",
    "RequirementEvaluator",
    "PlausibilityEvaluator",
    "CorrectnessEvaluator",
]



from typing import Optional
from typing_extensions import deprecated

import logging
import os

from .base import ChatMessage, LLMClient
from .logger import LLMLogger

logger = logging.getLogger(__name__)

_default_client = None

_default_llm_model = os.getenv("GSK_LLM_MODEL", "gpt-4o")
_disable_structured_output = False
_default_completion_params = dict()
_default_llm_api = None


if "GSK_LLM_BASE_URL" in os.environ is not None:
    _default_completion_params["api_base"] = os.getenv("GSK_LLM_BASE_URL")


@deprecated("set_default_client is deprecated: https://docs.giskard.ai/en/latest/open_source/setting_up/index.html")
def set_default_client(client: LLMClient):
    global _default_client
    _default_client = client


def _unset_default_client():
    global _default_client
    _default_client = None


def get_default_llm_api() -> str:
    global _default_llm_api
    if _default_llm_api is None:
        _default_llm_api = os.getenv(
            "GSK_LLM_API", "azure" if "AZURE_OPENAI_API_KEY" in os.environ else "openai"
        ).lower()

        if _default_llm_api not in {"azure", "openai"}:
            logging.warning(
                f"LLM-based evaluation is only working with `azure` and `openai`. Found {_default_llm_api} in GSK_LLM_API, falling back to `openai`"
            )
            _default_llm_api = "openai"

    return _default_llm_api


@deprecated("set_llm_api is deprecated: https://docs.giskard.ai/en/latest/open_source/setting_up/index.html")
def set_llm_api(llm_api: str):
    if llm_api.lower() not in {"azure", "openai"}:
        raise ValueError("Giskard LLM-based evaluators is only working with `azure` and `openai`")

    global _default_llm_api
    _default_llm_api = llm_api.lower()
    # If the API is set, we unset the default client
    _unset_default_client()


@deprecated("set_default_client is deprecated: https://docs.giskard.ai/en/latest/open_source/setting_up/index.html")
def set_llm_base_url(llm_base_url: Optional[str]):
    global _default_completion_params
    _default_completion_params["api_base"] = llm_base_url


def set_llm_model(llm_model: str, disable_structured_output=False, **kwargs):
    """

    :param llm_model: The model to be used
    :param disable_structured_output: Set this to True when the used model doesn't support structured output
    :param kwargs: Additional fixed params to be passed during completion
    """
    global _default_llm_model
    global _disable_structured_output
    global _default_completion_params

    _default_llm_model = llm_model
    _disable_structured_output = disable_structured_output
    _default_completion_params = kwargs

    # If the model is set, we unset the default client
    _unset_default_client()


def get_default_client() -> LLMClient:
    global _default_client
    global _default_llm_api
    global _default_llm_model
    global _disable_structured_output

    if _default_client is not None:
        return _default_client

    try:
        from .litellm import LiteLLMClient

        if (
            _default_llm_api is not None
            and "/" in _default_llm_model
            and not _default_llm_model.startswith(f"{_default_llm_api}/")
        ):
            raise ValueError(
                f"Model {_default_llm_model} is not compatible with {_default_llm_api}: https://docs.giskard.ai/en/latest/open_source/setting_up/index.html "
            )
        if _default_llm_api is not None and "/" not in _default_llm_model:
            _default_llm_model = f"{_default_llm_api}/{_default_llm_model}"

        _default_client = LiteLLMClient(_default_llm_model, _disable_structured_output, _default_completion_params)
    except ImportError:
        raise ValueError(f"LLM scan using {_default_llm_model} requires litellm")

    return _default_client


__all__ = [
    "LLMClient",
    "ChatMessage",
    "LLMLogger",
    "get_default_client",
    "set_llm_model",
    "get_default_llm_api",
    "set_llm_api",
]
